"""PhonopyAtoms class and routines related to atoms."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import re
import warnings
from collections.abc import Sequence
from math import gcd
from typing import Optional, Union

import numpy as np


def Atoms(*args, **kwargs):
    """Atoms class that is same as PhonopyAtoms class.

    This exists backward compatibility.

    """
    warnings.warn(
        "phonopy.atoms.Atoms is deprecated. Please use PhonopyAtoms instead of Atoms.",
        DeprecationWarning,
        stacklevel=2,
    )
    return PhonopyAtoms(*args, **kwargs)


def split_symbol_and_index(symnum: str):
    """Split symbol and index.

    H --> H, 0
    H2 --> H, 2

    """
    m = re.match(r"([a-zA-Z]+)([0-9]*)", symnum)
    symbol, index = m.groups()
    if symnum != f"{symbol}{index}":
        raise RuntimeError(f"Invalid symbol: {symnum}.")
    if index:
        index = int(index)
        if index < 1:
            raise RuntimeError(
                f"Invalid symbol. Index has to be greater than 0: {symnum}."
            )
    else:
        index = 0
    return symbol, index


class PhonopyAtoms:
    """Class to represent crystal structure.

    Originally this aimed to be compatible ASE Atoms class, but now not.

    Attributes
    ----------
    cell : np.ndarray
        Basis vectors (a, b, c) given in row vectors.
    positions : np.ndarray
        Positions of atoms in Cartesian coordinates. shape=(natom, 3),
        dtype='double', order='C'
    scaled_positions : np.ndarray
        Positions of atoms in fractional (crystallographic) coordinates.
        shape=(natom, 3), dtype='double', order='C'
    symbols : list[str]
        List of chemical symbols of atoms. Chemical symbol + natural number is
        allowed, e.g., "Cl1".
    numbers : np.ndarray
        Atomic numbers. Numbers cannot exceed 118. shape=(natom,), dtype='intc'
    masses : np.ndarray, optional
        Atomic masses. shape=(natom,), dtype='double'
    magnetic_moments : np.ndarray, optional
        shape=(natom,) or (natom, 3), dtype='double', order='C'
    volume : float
        Cell volume.
    Z : int
        Number of formula units in this cell.

    """

    _MOD_DIVISOR = 1000

    def __init__(
        self,
        symbols: Optional[Sequence] = None,
        numbers: Optional[Union[Sequence, np.ndarray]] = None,
        masses: Optional[Union[Sequence, np.ndarray]] = None,
        magnetic_moments: Optional[Union[Sequence, np.ndarray]] = None,
        scaled_positions: Optional[Union[Sequence, np.ndarray]] = None,
        positions: Optional[Union[Sequence, np.ndarray]] = None,
        cell: Optional[Union[Sequence, np.ndarray]] = None,
        atoms: Optional["PhonopyAtoms"] = None,
        magmoms: Optional[Union[Sequence, np.ndarray]] = None,
        pbc: Optional[bool] = None,
    ):  # pbc is dummy argument, and never used.
        """Init method."""
        if magmoms is not None:
            warnings.warn(
                "PhonopyAtoms.__init__ parameter of magmoms is deprecated. "
                "Use magnetic_moments instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        if pbc is not None:
            warnings.warn(
                "PhonopyAtoms.__init__ parameter of pbc is deprecated. "
                "It is considered always True.",
                DeprecationWarning,
                stacklevel=2,
            )
        if atoms:
            warnings.warn(
                "PhonopyAtoms.__init__ parameter of atoms is deprecated. "
                "Use PhonopyAtoms.copy() instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            self._set_parameters(
                numbers=atoms.numbers,
                masses=atoms.masses,
                magnetic_moments=atoms.magnetic_moments,
                scaled_positions=atoms.scaled_positions,
                cell=atoms.cell,
            )
        else:
            self._set_parameters(
                symbols=symbols,
                numbers=numbers,
                masses=masses,
                magnetic_moments=magnetic_moments,
                scaled_positions=scaled_positions,
                positions=positions,
                cell=cell,
            )

    def _set_parameters(
        self,
        symbols: Optional[Sequence] = None,
        numbers: Optional[Union[Sequence, np.ndarray]] = None,
        masses: Optional[Union[Sequence, np.ndarray]] = None,
        magnetic_moments: Optional[Union[Sequence, np.ndarray]] = None,
        scaled_positions: Optional[Union[Sequence, np.ndarray]] = None,
        positions: Optional[Union[Sequence, np.ndarray]] = None,
        cell: Optional[Union[Sequence, np.ndarray]] = None,
    ):
        self._cell = None
        self._scaled_positions = None
        self._set_cell_and_positions(
            cell, positions=positions, scaled_positions=scaled_positions
        )

        self._symbols = symbols

        self._numbers_with_shifts = None
        if numbers is not None:
            if (np.array(numbers) > 118).any():  # 118 is the max atomic number.
                raise RuntimeError("Atomic numbers cannot be larger than 118.")
            self._numbers_with_shifts = np.array(numbers, dtype="intc")

        self._masses = None
        self._set_masses(masses)

        # (initial) magnetic moments
        self._magnetic_moments = None
        self._set_magnetic_moments(magnetic_moments)

        # numbers <--> symbols
        if self._numbers_with_shifts is not None:  # number --> symbol
            self._numbers_to_symbols()
        elif self._symbols is not None:  # symbol --> number
            self._symbols_to_numbers()

        # symbol --> mass
        if self._symbols and (self._masses is None):
            if (self._numbers_with_shifts > 118).any():  # 118 is the max atomic number.
                raise RuntimeError(
                    "Masses have to be specified when special symbols are used."
                )
            self._symbols_to_masses()

        self._check()
        self._finalize()

    def __len__(self):
        """Return number of atoms."""
        return len(self.numbers)

    @property
    def cell(self):
        """Setter and getter of basis vectors. For getter, copy is returned."""
        return self._cell.copy()

    @cell.setter
    def cell(self, cell):
        self._set_cell(cell)
        self._check()

    def set_cell(self, cell):
        """Set basis vectors."""
        warnings.warn(
            "PhonopyAtoms.set_cell() is deprecated. Use cell attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.cell = cell

    def get_cell(self):
        """Return copy of basis vectors."""
        warnings.warn(
            "PhonopyAtoms.get_cell() is deprecated. Use cell attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.cell

    @property
    def positions(self):
        """Setter and getter of positions in Cartesian coordinates."""
        return np.array(
            np.dot(self._scaled_positions, self._cell), dtype="double", order="C"
        )

    @positions.setter
    def positions(self, positions):
        self._set_positions(positions)
        self._check()

    def get_positions(self):
        """Return positions in Cartesian coordinates."""
        warnings.warn(
            "PhonopyAtoms.get_positions() is deprecated. "
            "Use positions attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.positions

    def set_positions(self, positions):
        """Set positions in Cartesian coordinates."""
        warnings.warn(
            "PhonopyAtoms.set_positions() is deprecated. "
            "Use positions attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.positions = positions

    @property
    def scaled_positions(self):
        """Setter and getter of scaled positions. For getter, copy is returned."""
        return self._scaled_positions.copy()

    @scaled_positions.setter
    def scaled_positions(self, scaled_positions):
        self._set_scaled_positions(scaled_positions)
        self._check()

    def get_scaled_positions(self):
        """Return scaled positions."""
        warnings.warn(
            "PhonopyAtoms.get_scaled_positions() is deprecated. "
            "Use scaled_positions attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.scaled_positions

    def set_scaled_positions(self, scaled_positions):
        """Set scaled positions."""
        warnings.warn(
            "PhonopyAtoms.set_scaled_positions() is deprecated. "
            "Use scaled_positions attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.scaled_positions = scaled_positions

    @property
    def symbols(self):
        """Setter and getter of chemical symbols."""
        return self._symbols[:]

    @symbols.setter
    def symbols(self, symbols):
        warnings.warn(
            "Setter of PhonopyAtoms.symbols is deprecated.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._symbols = symbols
        self._check()
        self._symbols_to_numbers()
        self._symbols_to_masses()

    def get_chemical_symbols(self):
        """Return chemical symbols."""
        warnings.warn(
            "PhonopyAtoms.get_chemical_symbols() is deprecated. "
            "Use symbols attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.symbols

    def set_chemical_symbols(self, symbols):
        """Set chemical symbols."""
        warnings.warn(
            "PhonopyAtoms.set_chemical_symbols() is deprecated. "
            "Use symbols attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.symbols = symbols

    @property
    def numbers_with_shifts(self):
        """Getter of atomic numbers + MOD_DIVISOR * index."""
        return self._numbers_with_shifts.copy()

    @property
    def numbers(self):
        """Setter and getter of atomic numbers. For getter, new array is returned."""
        return np.array(
            [n % self._MOD_DIVISOR for n in self._numbers_with_shifts], dtype="intc"
        )

    @numbers.setter
    def numbers(self, numbers):
        if (np.array(numbers) > 118).any():  # 118 is the max atomic number.
            raise RuntimeError("Atomic number is too large.")
        warnings.warn(
            "Setter of PhonopyAtoms.number is deprecated.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._numbers_with_shifts = numbers
        self._check()
        self._numbers_to_symbols()
        self._symbols_to_masses()

    def get_atomic_numbers(self):
        """Return atomic numbers."""
        warnings.warn(
            "PhonopyAtoms.get_atomic_numbers() is deprecated. "
            "Use numbers attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.numbers

    def set_atomic_numbers(self, numbers):
        """Set atomic numbers."""
        warnings.warn(
            "PhonopyAtoms.set_atomic_numbers() is deprecated. "
            "Use numbers attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.numbers = numbers

    @property
    def masses(self):
        """Setter and getter of atomic masses. For getter copy is returned."""
        if self._masses is None:
            return None
        else:
            return self._masses.copy()

    @masses.setter
    def masses(self, masses):
        self._set_masses(masses)
        self._check()

    def get_masses(self):
        """Return atomic masses."""
        warnings.warn(
            "PhonopyAtoms.get_masses() is deprecated. Use masses attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.masses

    def set_masses(self, masses):
        """Set atomic masses."""
        warnings.warn(
            "PhonopyAtoms.set_masses() is deprecated. Use masses attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.masses = masses

    @property
    def magnetic_moments(self):
        """Setter and getter of magnetic moments. For getter, copy is returned.

        shape=(natom,) or (natom, 3), dtype='double', order='C'

        For setter, the formar can be specified by (natom, 1), which will be
        recognized as (natom,) and the latter can be specified by (natom * 3,),
        which will be converted to (natom, 3).

        """
        if self._magnetic_moments is None:
            return None
        else:
            return self._magnetic_moments.copy()

    @magnetic_moments.setter
    def magnetic_moments(self, magnetic_moments):
        self._set_magnetic_moments(magnetic_moments)
        self._check()
        self._finalize()

    def get_magnetic_moments(self):
        """Return magnetic moments."""
        warnings.warn(
            "PhonopyAtoms.get_magnetic_moments() is deprecated. "
            "Use magnetic_moments attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.magnetic_moments

    def set_magnetic_moments(self, magmoms):
        """Set magnetic moments."""
        warnings.warn(
            "PhonopyAtoms.set_magnetic_moments() is deprecated. "
            "Use magnetic_moments attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.magnetic_moments = magmoms

    @property
    def volume(self):
        """Return cell volume."""
        return np.linalg.det(self._cell)

    def get_volume(self):
        """Return cell volume."""
        warnings.warn(
            "PhonopyAtoms.get_volume() is deprecated. Use volume attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.volume

    @property
    def Z(self):
        """Return number of formula units in this cell."""
        count = {}
        for n in self._numbers_with_shifts:
            if n in count:
                count[n] += 1
            else:
                count[n] = 1
        values = list(count.values())
        x = values[0]
        for v in values[1:]:
            x = gcd(x, v)
        return x

    def get_number_of_atoms(self):
        """Return number of atoms."""
        warnings.warn(
            "PhonopyAtoms.get_number_of_atoms() is deprecated. Use len(PhonopyAtoms).",
            DeprecationWarning,
            stacklevel=2,
        )
        return len(self)

    def _set_cell(self, cell):
        _cell = np.array(cell, dtype="double", order="C")
        if _cell.shape == (3, 3):
            self._cell = _cell
        else:
            raise TypeError("Array shape of cell is not 3x3.")

    def _set_positions(self, cart_positions):
        self._scaled_positions = np.array(
            np.dot(cart_positions, np.linalg.inv(self._cell)), dtype="double", order="C"
        )

    def _set_scaled_positions(self, scaled_positions):
        self._scaled_positions = np.array(scaled_positions, dtype="double", order="C")

    def _set_masses(self, masses):
        if masses is None:
            self._masses = None
        else:
            self._masses = np.array(masses, dtype="double")

    def _set_magnetic_moments(self, magmoms):
        if magmoms is None:
            self._magnetic_moments = None
        else:
            self._magnetic_moments = np.array(np.ravel(magmoms), dtype="double")

    def _set_cell_and_positions(self, cell, positions=None, scaled_positions=None):
        self._set_cell(cell)
        if positions is not None:
            self._set_positions(positions)
        elif scaled_positions is not None:
            self._set_scaled_positions(scaled_positions)

    def _numbers_to_symbols(self):
        symbols = []
        for number in self._numbers_with_shifts:
            n = number % self._MOD_DIVISOR
            m = number // self._MOD_DIVISOR
            if m > 0:
                symbols.append(f"{atom_data[n][1]}{m}")
            else:
                symbols.append(f"{atom_data[n][1]}")
        self._symbols = symbols

    def _symbols_to_numbers(self):
        numbers = []
        for symnum in self._symbols:
            symbol, index = split_symbol_and_index(symnum)
            numbers.append(symbol_map[symbol] + self._MOD_DIVISOR * index)

        self._numbers_with_shifts = np.array(numbers, dtype="intc")

    def _symbols_to_masses(self):
        masses = [atom_data[symbol_map[s]][3] for s in self._symbols]
        if None in masses:
            self._masses = None
        else:
            self._masses = np.array(masses, dtype="double")

    def _check(self):
        """Check number of eleemnts in arrays.

        Do not modify the arrays. Modification of array shapes should be done in
        ``self._finalize()``.

        """
        if self._cell is None:
            raise RuntimeError("cell is not set.")
        if self._scaled_positions is None:
            raise RuntimeError("scaled_positions (positions) is not set.")
        if self._numbers_with_shifts is None:
            raise RuntimeError("numbers is not set.")
        if len(self._numbers_with_shifts) != len(self._scaled_positions):
            raise RuntimeError("len(numbers) != len(scaled_positions).")
        if len(self._numbers_with_shifts) != len(self._symbols):
            raise RuntimeError("len(numbers) != len(symbols).")
        if self._masses is not None:
            if len(self._numbers_with_shifts) != len(self._masses):
                raise RuntimeError("len(numbers) != len(masses).")
        if self._magnetic_moments is not None:
            if len(self._magnetic_moments.ravel()) not in (len(self), len(self) * 3):
                raise RuntimeError(
                    "magnetic_moments has to have shape=(natom,) or (natom, 3)."
                )

    def _finalize(self):
        """Modify array shapes to those expeted to be exposed."""
        # When non collinear magnetic moments is given in a flat array.
        if self.magnetic_moments is not None:
            if len(self.magnetic_moments.ravel()) == len(self) * 3:
                self._magnetic_moments = np.reshape(self._magnetic_moments, (-1, 3))

    def copy(self):
        """Return copy of itself."""
        return PhonopyAtoms(
            cell=self._cell,
            scaled_positions=self._scaled_positions,
            masses=self._masses,
            magnetic_moments=self._magnetic_moments,
            symbols=self._symbols,
        )

    def totuple(self, distinguish_symbol_index: bool = False):
        """Return (cell, scaled_position, numbers).

        If magmams is set, (cell, scaled_position, numbers, magmoms) is
        returned.

        Parameters
        ----------
        with_symbol_index : bool
            If True, number is replaced with atomic number + index *
            self.MOD_DIVISOR.

            'H' --> 1
            'H2' --> 1 + self.MOD_DIVISOR * 2

        """
        if distinguish_symbol_index:
            numbers = self._numbers_with_shifts
        else:
            numbers = self.numbers

        if self._magnetic_moments is None:
            return (self._cell, self._scaled_positions, numbers)
        else:
            return (
                self._cell,
                self._scaled_positions,
                numbers,
                self._magnetic_moments,
            )

    def to_tuple(self):
        """Return (cell, scaled_position, numbers).

        If magmams is set, (cell, scaled_position, numbers, magmoms) is returned.

        """
        warnings.warn(
            "PhonopyAtoms.to_tuple() is deprecated. Use totuple() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.totuple()

    def get_yaml_lines(self):
        """Return list of text lines of crystal structure in yaml."""
        lines = ["lattice:"]
        for v, a in zip(self._cell, ("a", "b", "c")):
            lines.append("- [ %21.15f, %21.15f, %21.15f ] # %s" % (v[0], v[1], v[2], a))
        lines.append("points:")
        if self._masses is None:
            masses = [None] * len(self._symbols)
        else:
            masses = self._masses
        if self._magnetic_moments is None:
            magmoms = [None] * len(self._symbols)
        else:
            magmoms = self._magnetic_moments
        for i, (s, num, v, m, mag) in enumerate(
            zip(self._symbols, self.numbers, self._scaled_positions, masses, magmoms)
        ):
            formal_s = atom_data[num][1]
            if s == formal_s:
                lines.append(f"- symbol: {s} # {i + 1}")
            else:
                lines.append(f"- symbol: {formal_s} # {i + 1}")
                lines.append(f"  extended_symbol: {s}")
            lines.append("  coordinates: [ %18.15f, %18.15f, %18.15f ]" % tuple(v))
            if m is not None:
                lines.append("  mass: %f" % m)
            if mag is not None:
                if mag.ndim == 0:
                    mag_str = f"{mag:.8f}"
                else:
                    mag_str = f"[{mag[0]:.8f}, {mag[1]:.8f}, {mag[2]:.8f}]"
                lines.append(f"  magnetic_moment: {mag_str}")
        return lines

    def __str__(self):
        """Return text lines of crystal structure in yaml."""
        return "\n".join(self.get_yaml_lines())

    def _get_element_counts(self):
        """Return dict of element counts, with indices stripped from symbols."""
        counts = {}
        for symbol in self._symbols:
            base_symbol = symbol.rstrip("0123456789")
            counts[base_symbol] = counts.get(base_symbol, 0) + 1
        return counts

    def _build_formula(self, counts, divisor=1):
        """Build formula string from element counts and optional divisor.

        Parameters
        ----------
        counts : dict
            Dictionary mapping element symbols to their counts
        divisor : int, optional
            Number to divide counts by, defaults to 1

        Returns
        -------
        str
            Formula string with elements in alphabetical order
        """
        if not counts:
            return ""

        formula_parts = []
        for element in sorted(counts):
            count = counts[element] // divisor
            if count == 1:
                formula_parts.append(element)
            else:
                formula_parts.append(f"{element}{count}")

        return "".join(formula_parts)

    @property
    def formula(self):
        """Return chemical formula as a string.

        The formula is constructed by sorting elements alphabetically and
        appending numbers for elements that appear more than once.
        E.g., "Si2O4" for two Si and four O atoms.
        """
        counts = self._get_element_counts()
        return self._build_formula(counts)

    @property
    def reduced_formula(self):
        """Return reduced chemical formula as a string.

        The reduced formula divides all element counts by their GCD.
        E.g., "Fe4O8" becomes "Fe2O4".
        """
        counts = self._get_element_counts()
        if not counts:
            return ""

        # Find GCD of all counts
        numbers = list(counts.values())
        divisor = numbers[0]
        for n in numbers[1:]:
            divisor = gcd(divisor, n)

        return self._build_formula(counts, divisor)

    @property
    def normalized_formula(self):
        """Return normalized formula as a string.

        The normalized formula scales all element counts so they sum to 1.
        E.g., "Fe2O3" becomes "Fe0.4O0.6".
        """
        counts = self._get_element_counts()
        if not counts:
            return ""

        # Get total count
        total = sum(counts.values())

        # Build normalized formula string
        formula_parts = []
        for element in sorted(counts):
            count = counts[element] / total
            # Always show decimal for normalized formula
            formula_parts.append(f"{element}{count:.3}")

        return "".join(formula_parts)


def parse_cell_dict(cell_dict: dict) -> Optional[PhonopyAtoms]:
    """Parse cell dict."""
    lattice = None
    if "lattice" in cell_dict:
        lattice = cell_dict["lattice"]
    else:
        return None
    points = []
    symbols = []
    masses = []
    magnetic_moments = []
    if "points" in cell_dict:
        for x in cell_dict["points"]:
            if "coordinates" in x:
                points.append(x["coordinates"])
            if "extended_symbol" in x:  # like Fe1
                symbols.append(x["extended_symbol"])
            elif "symbol" in x:  # like Fe
                symbols.append(x["symbol"])
            if "mass" in x:
                masses.append(x["mass"])
            if "magnetic_moment" in x:
                magnetic_moments.append(x["magnetic_moment"])
    # For version < 1.10.9
    elif "atoms" in cell_dict:
        for x in cell_dict["atoms"]:
            if "coordinates" not in x and "position" in x:
                points.append(x["position"])
            if "symbol" in x:
                symbols.append(x["symbol"])
            if "mass" in x:
                masses.append(x["mass"])

    if not masses:
        masses = None
    if not magnetic_moments:
        magnetic_moments = None

    if points and symbols:
        return PhonopyAtoms(
            symbols=symbols,
            cell=lattice,
            masses=masses,
            scaled_positions=points,
            magnetic_moments=magnetic_moments,
        )
    else:
        return None


# Pure Appl. Chem., Vol. 83, No. 2, pp. 359-396, 2011. is available
# but the following list is from 2006.

# Pure Appl. Chem., Vol. 78, No. 11, pp. 2051-2066, 2006.
# The masses of following elements are obtained from wikipedia:
# Ac: 227
# Np: 237
# Pm: 145
# Tc: 98
atom_data = [
    [0, "X", "X", None],  # 0
    [1, "H", "Hydrogen", 1.00794],  # 1
    [2, "He", "Helium", 4.002602],  # 2
    [3, "Li", "Lithium", 6.941],  # 3
    [4, "Be", "Beryllium", 9.012182],  # 4
    [5, "B", "Boron", 10.811],  # 5
    [6, "C", "Carbon", 12.0107],  # 6
    [7, "N", "Nitrogen", 14.0067],  # 7
    [8, "O", "Oxygen", 15.9994],  # 8
    [9, "F", "Fluorine", 18.9984032],  # 9
    [10, "Ne", "Neon", 20.1797],  # 10
    [11, "Na", "Sodium", 22.98976928],  # 11
    [12, "Mg", "Magnesium", 24.3050],  # 12
    [13, "Al", "Aluminium", 26.9815386],  # 13
    [14, "Si", "Silicon", 28.0855],  # 14
    [15, "P", "Phosphorus", 30.973762],  # 15
    [16, "S", "Sulfur", 32.065],  # 16
    [17, "Cl", "Chlorine", 35.453],  # 17
    [18, "Ar", "Argon", 39.948],  # 18
    [19, "K", "Potassium", 39.0983],  # 19
    [20, "Ca", "Calcium", 40.078],  # 20
    [21, "Sc", "Scandium", 44.955912],  # 21
    [22, "Ti", "Titanium", 47.867],  # 22
    [23, "V", "Vanadium", 50.9415],  # 23
    [24, "Cr", "Chromium", 51.9961],  # 24
    [25, "Mn", "Manganese", 54.938045],  # 25
    [26, "Fe", "Iron", 55.845],  # 26
    [27, "Co", "Cobalt", 58.933195],  # 27
    [28, "Ni", "Nickel", 58.6934],  # 28
    [29, "Cu", "Copper", 63.546],  # 29
    [30, "Zn", "Zinc", 65.38],  # 30
    [31, "Ga", "Gallium", 69.723],  # 31
    [32, "Ge", "Germanium", 72.64],  # 32
    [33, "As", "Arsenic", 74.92160],  # 33
    [34, "Se", "Selenium", 78.96],  # 34
    [35, "Br", "Bromine", 79.904],  # 35
    [36, "Kr", "Krypton", 83.798],  # 36
    [37, "Rb", "Rubidium", 85.4678],  # 37
    [38, "Sr", "Strontium", 87.62],  # 38
    [39, "Y", "Yttrium", 88.90585],  # 39
    [40, "Zr", "Zirconium", 91.224],  # 40
    [41, "Nb", "Niobium", 92.90638],  # 41
    [42, "Mo", "Molybdenum", 95.96],  # 42
    [43, "Tc", "Technetium", 98],  # 43 (mass is from wikipedia)
    [44, "Ru", "Ruthenium", 101.07],  # 44
    [45, "Rh", "Rhodium", 102.90550],  # 45
    [46, "Pd", "Palladium", 106.42],  # 46
    [47, "Ag", "Silver", 107.8682],  # 47
    [48, "Cd", "Cadmium", 112.411],  # 48
    [49, "In", "Indium", 114.818],  # 49
    [50, "Sn", "Tin", 118.710],  # 50
    [51, "Sb", "Antimony", 121.760],  # 51
    [52, "Te", "Tellurium", 127.60],  # 52
    [53, "I", "Iodine", 126.90447],  # 53
    [54, "Xe", "Xenon", 131.293],  # 54
    [55, "Cs", "Caesium", 132.9054519],  # 55
    [56, "Ba", "Barium", 137.327],  # 56
    [57, "La", "Lanthanum", 138.90547],  # 57
    [58, "Ce", "Cerium", 140.116],  # 58
    [59, "Pr", "Praseodymium", 140.90765],  # 59
    [60, "Nd", "Neodymium", 144.242],  # 60
    [61, "Pm", "Promethium", 145],  # 61 (mass is from wikipedia)
    [62, "Sm", "Samarium", 150.36],  # 62
    [63, "Eu", "Europium", 151.964],  # 63
    [64, "Gd", "Gadolinium", 157.25],  # 64
    [65, "Tb", "Terbium", 158.92535],  # 65
    [66, "Dy", "Dysprosium", 162.500],  # 66
    [67, "Ho", "Holmium", 164.93032],  # 67
    [68, "Er", "Erbium", 167.259],  # 68
    [69, "Tm", "Thulium", 168.93421],  # 69
    [70, "Yb", "Ytterbium", 173.054],  # 70
    [71, "Lu", "Lutetium", 174.9668],  # 71
    [72, "Hf", "Hafnium", 178.49],  # 72
    [73, "Ta", "Tantalum", 180.94788],  # 73
    [74, "W", "Tungsten", 183.84],  # 74
    [75, "Re", "Rhenium", 186.207],  # 75
    [76, "Os", "Osmium", 190.23],  # 76
    [77, "Ir", "Iridium", 192.217],  # 77
    [78, "Pt", "Platinum", 195.084],  # 78
    [79, "Au", "Gold", 196.966569],  # 79
    [80, "Hg", "Mercury", 200.59],  # 80
    [81, "Tl", "Thallium", 204.3833],  # 81
    [82, "Pb", "Lead", 207.2],  # 82
    [83, "Bi", "Bismuth", 208.98040],  # 83
    [84, "Po", "Polonium", None],  # 84
    [85, "At", "Astatine", None],  # 85
    [86, "Rn", "Radon", None],  # 86
    [87, "Fr", "Francium", None],  # 87
    [88, "Ra", "Radium", None],  # 88
    [89, "Ac", "Actinium", 227],  # 89 (mass is from wikipedia)
    [90, "Th", "Thorium", 232.03806],  # 90
    [91, "Pa", "Protactinium", 231.03588],  # 91
    [92, "U", "Uranium", 238.02891],  # 92
    [93, "Np", "Neptunium", 237],  # 93 (mass is from wikipedia)
    [94, "Pu", "Plutonium", None],  # 94
    [95, "Am", "Americium", None],  # 95
    [96, "Cm", "Curium", None],  # 96
    [97, "Bk", "Berkelium", None],  # 97
    [98, "Cf", "Californium", None],  # 98
    [99, "Es", "Einsteinium", None],  # 99
    [100, "Fm", "Fermium", None],  # 100
    [101, "Md", "Mendelevium", None],  # 101
    [102, "No", "Nobelium", None],  # 102
    [103, "Lr", "Lawrencium", None],  # 103
    [104, "Rf", "Rutherfordium", None],  # 104
    [105, "Db", "Dubnium", None],  # 105
    [106, "Sg", "Seaborgium", None],  # 106
    [107, "Bh", "Bohrium", None],  # 107
    [108, "Hs", "Hassium", None],  # 108
    [109, "Mt", "Meitnerium", None],  # 109
    [110, "Ds", "Darmstadtium", None],  # 110
    [111, "Rg", "Roentgenium", None],  # 111
    [112, "Cn", "Copernicium", None],  # 112
    [113, "Uut", "Ununtrium", None],  # 113
    [114, "Uuq", "Ununquadium", None],  # 114
    [115, "Uup", "Ununpentium", None],  # 115
    [116, "Uuh", "Ununhexium", None],  # 116
    [117, "Uus", "Ununseptium", None],  # 117
    [118, "Uuo", "Ununoctium", None],  # 118
]

symbol_map = {
    "H": 1,
    "He": 2,
    "Li": 3,
    "Be": 4,
    "B": 5,
    "C": 6,
    "N": 7,
    "O": 8,
    "F": 9,
    "Ne": 10,
    "Na": 11,
    "Mg": 12,
    "Al": 13,
    "Si": 14,
    "P": 15,
    "S": 16,
    "Cl": 17,
    "Ar": 18,
    "K": 19,
    "Ca": 20,
    "Sc": 21,
    "Ti": 22,
    "V": 23,
    "Cr": 24,
    "Mn": 25,
    "Fe": 26,
    "Co": 27,
    "Ni": 28,
    "Cu": 29,
    "Zn": 30,
    "Ga": 31,
    "Ge": 32,
    "As": 33,
    "Se": 34,
    "Br": 35,
    "Kr": 36,
    "Rb": 37,
    "Sr": 38,
    "Y": 39,
    "Zr": 40,
    "Nb": 41,
    "Mo": 42,
    "Tc": 43,
    "Ru": 44,
    "Rh": 45,
    "Pd": 46,
    "Ag": 47,
    "Cd": 48,
    "In": 49,
    "Sn": 50,
    "Sb": 51,
    "Te": 52,
    "I": 53,
    "Xe": 54,
    "Cs": 55,
    "Ba": 56,
    "La": 57,
    "Ce": 58,
    "Pr": 59,
    "Nd": 60,
    "Pm": 61,
    "Sm": 62,
    "Eu": 63,
    "Gd": 64,
    "Tb": 65,
    "Dy": 66,
    "Ho": 67,
    "Er": 68,
    "Tm": 69,
    "Yb": 70,
    "Lu": 71,
    "Hf": 72,
    "Ta": 73,
    "W": 74,
    "Re": 75,
    "Os": 76,
    "Ir": 77,
    "Pt": 78,
    "Au": 79,
    "Hg": 80,
    "Tl": 81,
    "Pb": 82,
    "Bi": 83,
    "Po": 84,
    "At": 85,
    "Rn": 86,
    "Fr": 87,
    "Ra": 88,
    "Ac": 89,
    "Th": 90,
    "Pa": 91,
    "U": 92,
    "Np": 93,
    "Pu": 94,
    "Am": 95,
    "Cm": 96,
    "Bk": 97,
    "Cf": 98,
    "Es": 99,
    "Fm": 100,
    "Md": 101,
    "No": 102,
    "Lr": 103,
    "Rf": 104,
    "Db": 105,
    "Sg": 106,
    "Bh": 107,
    "Hs": 108,
    "Mt": 109,
    "Ds": 110,
    "Rg": 111,
    "Cn": 112,
    "Uut": 113,
    "Uuq": 114,
    "Uup": 115,
    "Uuh": 116,
    "Uus": 117,
    "Uuo": 118,
}


isotope_data_CIAAW = {
    # https://www.ciaaw.org/molybdenum.htm (accessed at 14th Jun. 2022)
    # Isotope Mo
    # Adam J. Mayer* and Michael E. Wieser, J. Anal. At. Spectrom., 2014, 29, 85
    # DOI: 10.1039/c3ja50164g
    # "The absolute isotopic composition and atomic weight of molybdenum in
    # SRM 3134 using an isotopic double-spike"
    "Mo": [
        [92, 91.906807, 0.14649],
        [94, 93.905084, 0.09187],
        [95, 94.9058374, 0.15873],
        [96, 95.9046748, 0.16673],
        [97, 96.906017, 0.09582],
        [98, 97.905404, 0.24292],
        [100, 99.907468, 0.09744],
    ]
}

# This data are obtained from
# J. R. de Laeter, J. K. Böhlke, P. De Bièvre, H. Hidaka, H. S. Peiser,
# K. J. R. Rosman and P. D. P. Taylor (2003).
# "Atomic weights of the elements. Review 2000 (IUPAC Technical Report)"
isotope_data = {
    "H": [[1, 1.0078250319, 0.999885], [2, 2.0141017779, 0.000115]],
    "He": [[3, 3.0160293094, 0.00000134], [4, 4.0026032497, 0.99999866]],
    "Li": [[6, 6.0151223, 0.0759], [7, 7.0160041, 0.9241]],
    "Be": [[9, 9.0121822, 1.0000]],
    "B": [[10, 10.0129371, 0.199], [11, 11.0093055, 0.801]],
    "C": [[12, 12, 0.9893], [13, 13.003354838, 0.0107]],
    "N": [[14, 14.0030740074, 0.99636], [15, 15.000108973, 0.00364]],
    "O": [
        [16, 15.9949146223, 0.99757],
        [17, 16.99913150, 0.00038],
        [18, 17.9991604, 0.00205],
    ],
    "F": [[19, 18.99840320, 1.0000]],
    "Ne": [
        [20, 19.992440176, 0.9048],
        [21, 20.99384674, 0.0027],
        [22, 21.99138550, 0.0925],
    ],
    "Na": [[23, 22.98976966, 1.0000]],
    "Mg": [
        [24, 23.98504187, 0.7899],
        [25, 24.98583700, 0.1000],
        [26, 25.98259300, 0.1101],
    ],
    "Al": [[27, 26.98153841, 1.0000]],
    "Si": [
        [28, 27.97692649, 0.92223],
        [29, 28.97649468, 0.04685],
        [30, 29.97377018, 0.03092],
    ],
    "P": [[31, 30.97376149, 1.0000]],
    "S": [
        [32, 31.97207073, 0.9499],
        [33, 32.97145854, 0.0075],
        [34, 33.96786687, 0.0425],
        [36, 35.96708088, 0.0001],
    ],
    "Cl": [[35, 34.96885271, 0.7576], [37, 36.96590260, 0.2424]],
    "Ar": [
        [36, 35.96754626, 0.003365],
        [38, 37.9627322, 0.000632],
        [40, 39.962383124, 0.996003],
    ],
    "K": [
        [39, 38.96370, 0.932581],
        [40, 39.96399867, 0.000117],
        [41, 40.96182597, 0.067302],
    ],
    "Ca": [
        [40, 39.9625912, 0.96941],
        [42, 41.9586183, 0.00647],
        [43, 42.9587668, 0.00135],
        [44, 43.9554811, 0.02086],
        [46, 45.9536927, 0.00004],
        [48, 47.952533, 0.00187],
    ],
    "Sc": [[45, 44.9559102, 1.0000]],
    "Ti": [
        [46, 45.9526295, 0.0825],
        [47, 46.9517637, 0.0744],
        [48, 47.9479470, 0.7372],
        [49, 48.9478707, 0.0541],
        [50, 49.9447920, 0.0518],
    ],
    "V": [[50, 49.9471627, 0.00250], [51, 50.9439635, 0.99750]],
    "Cr": [
        [50, 49.9460495, 0.04345],
        [52, 51.9405115, 0.83789],
        [53, 52.9406534, 0.09501],
        [54, 53.9388846, 0.02365],
    ],
    "Mn": [[55, 54.9380493, 1.0000]],
    "Fe": [
        [54, 53.9396147, 0.05845],
        [56, 55.9349418, 0.91754],
        [57, 56.9353983, 0.02119],
        [58, 57.9332801, 0.00282],
    ],
    "Co": [[59, 58.9331999, 1.0000]],
    "Ni": [
        [58, 57.9353477, 0.680769],
        [60, 59.9307903, 0.262231],
        [61, 60.9310601, 0.011399],
        [62, 61.9283484, 0.036345],
        [64, 63.9279692, 0.009256],
    ],
    "Cu": [[63, 62.9296007, 0.6915], [65, 64.9277938, 0.3085]],
    "Zn": [
        [64, 63.9291461, 0.48268],
        [66, 65.9260364, 0.27975],
        [67, 66.9271305, 0.04102],
        [68, 67.9248473, 0.19024],
        [70, 69.925325, 0.00631],
    ],
    "Ga": [[69, 68.925581, 0.60108], [71, 70.9247073, 0.39892]],
    "Ge": [
        [70, 69.9242500, 0.2038],
        [72, 71.9220763, 0.2731],
        [73, 72.9234595, 0.0776],
        [74, 73.9211784, 0.3672],
        [76, 75.921402, 0.0783],
    ],
    "As": [[75, 74.9215966, 1.0000]],
    "Se": [
        [74, 73.9224767, 0.0089],
        [76, 75.9192143, 0.0937],
        [77, 76.9199148, 0.0763],
        [78, 77.9173097, 0.2377],
        [80, 79.9165221, 0.4961],
        [82, 81.9167003, 0.0873],
    ],
    "Br": [[79, 78.9183379, 0.5069], [81, 80.916291, 0.4931]],
    "Kr": [
        [78, 77.920388, 0.00355],
        [80, 79.916379, 0.02286],
        [82, 81.9134850, 0.11593],
        [83, 82.914137, 0.11500],
        [84, 83.911508, 0.56987],
        [86, 85.910615, 0.17279],
    ],
    "Rb": [[85, 84.9117924, 0.7217], [87, 86.9091858, 0.2783]],
    "Sr": [
        [84, 83.913426, 0.0056],
        [86, 85.9092647, 0.0986],
        [87, 86.9088816, 0.0700],
        [88, 87.9056167, 0.8258],
    ],
    "Y": [[89, 88.9058485, 1.0000]],
    "Zr": [
        [90, 89.9047022, 0.5145],
        [91, 90.9056434, 0.1122],
        [92, 91.9050386, 0.1715],
        [94, 93.9063144, 0.1738],
        [96, 95.908275, 0.0280],
    ],
    "Nb": [[93, 92.9063762, 1.0000]],
    "Mo": [
        [92, 91.906810, 0.1477],
        [94, 93.9050867, 0.0923],
        [95, 94.9058406, 0.1590],
        [96, 95.9046780, 0.1668],
        [97, 96.9060201, 0.0956],
        [98, 97.905406, 0.2419],
        [100, 99.907476, 0.0967],
    ],
    "Tc": None,
    "Ru": [
        [96, 95.907604, 0.0554],
        [98, 97.905287, 0.0187],
        [99, 98.9059385, 0.1276],
        [100, 99.9042189, 0.1260],
        [101, 100.9055815, 0.1706],
        [102, 101.9043488, 0.3155],
        [104, 103.905430, 0.1862],
    ],
    "Rh": [[103, 102.905504, 1.0000]],
    "Pd": [
        [102, 101.905607, 0.0102],
        [104, 103.904034, 0.1114],
        [105, 104.905083, 0.2233],
        [106, 105.903484, 0.2733],
        [108, 107.903895, 0.2646],
        [110, 109.905153, 0.1172],
    ],
    "Ag": [[107, 106.905093, 0.51839], [109, 108.904756, 0.48161]],
    "Cd": [
        [106, 105.906458, 0.0125],
        [108, 107.904183, 0.0089],
        [110, 109.903006, 0.1249],
        [111, 110.904182, 0.1280],
        [112, 111.9027577, 0.2413],
        [113, 112.9044014, 0.1222],
        [114, 113.9033586, 0.2873],
        [116, 115.904756, 0.0749],
    ],
    "In": [[113, 112.904062, 0.0429], [115, 114.903879, 0.9571]],
    "Sn": [
        [112, 111.904822, 0.0097],
        [114, 113.902783, 0.0066],
        [115, 114.903347, 0.0034],
        [116, 115.901745, 0.1454],
        [117, 116.902955, 0.0768],
        [118, 117.901608, 0.2422],
        [119, 118.903311, 0.0859],
        [120, 119.9021985, 0.3258],
        [122, 121.9034411, 0.0463],
        [124, 123.9052745, 0.0579],
    ],
    "Sb": [[121, 120.9038222, 0.5721], [123, 122.9042160, 0.4279]],
    "Te": [
        [120, 119.904026, 0.0009],
        [122, 121.9030558, 0.0255],
        [123, 122.9042711, 0.0089],
        [124, 123.9028188, 0.0474],
        [125, 124.9044241, 0.0707],
        [126, 125.9033049, 0.1884],
        [128, 127.9044615, 0.3174],
        [130, 129.9062229, 0.3408],
    ],
    "I": [[127, 126.904468, 1.0000]],
    "Xe": [
        [124, 123.9058954, 0.000952],
        [126, 125.904268, 0.000890],
        [128, 127.9035305, 0.019102],
        [129, 128.9047799, 0.264006],
        [130, 129.9035089, 0.040710],
        [131, 130.9050828, 0.212324],
        [132, 131.9041546, 0.269086],
        [134, 133.9053945, 0.104357],
        [136, 135.907220, 0.088573],
    ],
    "Cs": [[133, 132.905447, 1.0000]],
    "Ba": [
        [130, 129.906311, 0.00106],
        [132, 131.905056, 0.00101],
        [134, 133.904504, 0.02417],
        [135, 134.905684, 0.06592],
        [136, 135.904571, 0.07854],
        [137, 136.905822, 0.11232],
        [138, 137.905242, 0.71698],
    ],
    "La": [[138, 137.907108, 0.00090], [139, 138.906349, 0.99910]],
    "Ce": [
        [136, 135.907140, 0.00185],
        [138, 137.905986, 0.00251],
        [140, 139.905435, 0.88450],
        [142, 141.909241, 0.11114],
    ],
    "Pr": [[141, 140.907648, 1.0000]],
    "Nd": [
        [142, 141.907719, 0.272],
        [143, 142.909810, 0.122],
        [144, 143.910083, 0.238],
        [145, 144.912569, 0.083],
        [146, 145.913113, 0.172],
        [148, 147.916889, 0.057],
        [150, 149.920887, 0.056],
    ],
    "Pm": None,
    "Sm": [
        [144, 143.911996, 0.0307],
        [147, 146.914894, 0.1499],
        [148, 147.914818, 0.1124],
        [149, 148.917180, 0.1382],
        [150, 149.917272, 0.0738],
        [152, 151.919729, 0.2675],
        [154, 153.922206, 0.2275],
    ],
    "Eu": [[151, 150.919846, 0.4781], [153, 152.921227, 0.5219]],
    "Gd": [
        [152, 151.919789, 0.0020],
        [154, 153.920862, 0.0218],
        [155, 154.922619, 0.1480],
        [156, 155.922120, 0.2047],
        [157, 156.923957, 0.1565],
        [158, 157.924101, 0.2484],
        [160, 159.927051, 0.2186],
    ],
    "Tb": [[159, 158.925343, 1.0000]],
    "Dy": [
        [156, 155.924278, 0.00056],
        [158, 157.924405, 0.00095],
        [160, 159.925194, 0.02329],
        [161, 160.926930, 0.18889],
        [162, 161.926795, 0.25475],
        [163, 162.928728, 0.24896],
        [164, 163.929171, 0.28260],
    ],
    "Ho": [[165, 164.930319, 1.0000]],
    "Er": [
        [162, 161.928775, 0.00139],
        [164, 163.929197, 0.01601],
        [166, 165.930290, 0.33503],
        [167, 166.932046, 0.22869],
        [168, 167.932368, 0.26978],
        [170, 169.935461, 0.14910],
    ],
    "Tm": [[169, 168.934211, 1.0000]],
    "Yb": [
        [168, 167.933895, 0.0013],
        [170, 169.934759, 0.0304],
        [171, 170.936323, 0.1428],
        [172, 171.936378, 0.2183],
        [173, 172.938207, 0.1613],
        [174, 173.938858, 0.3183],
        [176, 175.942569, 0.1276],
    ],
    "Lu": [[175, 174.9407682, 0.9741], [176, 175.9426827, 0.0259]],
    "Hf": [
        [174, 173.940042, 0.0016],
        [176, 175.941403, 0.0526],
        [177, 176.9432204, 0.1860],
        [178, 177.9436981, 0.2728],
        [179, 178.9458154, 0.1362],
        [180, 179.9465488, 0.3508],
    ],
    "Ta": [[180, 179.947466, 0.00012], [181, 180.947996, 0.99988]],
    "W": [
        [180, 179.946706, 0.0012],
        [182, 181.948205, 0.2650],
        [183, 182.9502242, 0.1431],
        [184, 183.9509323, 0.3064],
        [186, 185.95436, 0.2843],
    ],
    "Re": [[185, 184.952955, 0.3740], [187, 186.9557505, 0.6260]],
    "Os": [
        [184, 183.952491, 0.0002],
        [186, 185.953838, 0.0159],
        [187, 186.9557476, 0.0196],
        [188, 187.9558357, 0.1324],
        [189, 188.958145, 0.1615],
        [190, 189.958445, 0.2626],
        [192, 191.961479, 0.4078],
    ],
    "Ir": [[191, 190.960591, 0.373], [193, 192.962923, 0.627]],
    "Pt": [
        [190, 189.959930, 0.00014],
        [192, 191.961035, 0.00782],
        [194, 193.962663, 0.32967],
        [195, 194.964774, 0.33832],
        [196, 195.964934, 0.25242],
        [198, 197.967875, 0.07163],
    ],
    "Au": [[197, 196.966551, 1.0000]],
    "Hg": [
        [196, 195.965814, 0.0015],
        [198, 197.966752, 0.0997],
        [199, 198.968262, 0.1687],
        [200, 199.968309, 0.2310],
        [201, 200.970285, 0.1318],
        [202, 201.970625, 0.2986],
        [204, 203.973475, 0.0687],
    ],
    "Tl": [[203, 202.972329, 0.2952], [205, 204.974412, 0.7048]],
    "Pb": [
        [204, 203.973028, 0.014],
        [206, 205.974449, 0.241],
        [207, 206.975880, 0.221],
        [208, 207.976636, 0.524],
    ],
    "Bi": [[209, 208.980384, 1.0000]],
    "Po": None,
    "At": None,
    "Rn": None,
    "Fr": None,
    "Ra": None,
    "Ac": None,
    "Th": [[232, 232.0380495, 1.0000]],
    "Pa": [[231, 231.03588, 1.0000]],
    "U": [
        [234, 234.0409447, 0.000054],
        [235, 235.0439222, 0.007204],
        [238, 238.0507835, 0.992742],
    ],
}



# Copyright (C) 2012 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import warnings

import numpy as np

from phonopy.interface.vasp import get_born_OUTCAR, get_born_vasprunxml


def fracval(frac):
    """Convert fractional value string to float."""
    if frac.find("/") == -1:
        return float(frac)
    else:
        x = frac.split("/")
        return float(x[0]) / float(x[1])


def get_options():
    """Parse command-line options."""
    import argparse

    parser = argparse.ArgumentParser(description="Phonopy vasp-born command-line-tool")
    parser.set_defaults(
        num_atoms=None,
        primitive_axes=None,
        supercell_matrix=None,
        symmetrize_tensors=True,
        read_outcar=False,
        symprec=1e-5,
    )
    parser.add_argument(
        "--dim", dest="supercell_matrix", help="Same behavior as DIM tag"
    )
    parser.add_argument(
        "--pa",
        "--primitive-axis",
        "--primitive-axes",
        dest="primitive_axes",
        help="Same as PRIMITIVE_AXES tags",
    )
    parser.add_argument(
        "--nost",
        "--no-symmetrize-tensors",
        dest="symmetrize_tensors",
        action="store_false",
        help="Prevent from symmetrizing tensors",
    )
    parser.add_argument(
        "--tolerance", dest="symprec", type=float, help="Symmetry tolerance to search"
    )
    parser.add_argument(
        "--outcar",
        dest="read_outcar",
        action="store_true",
        help=("Read OUTCAR instead of vasprun.xml. POSCAR is necessary in this case."),
    )
    parser.add_argument(
        "filenames", nargs="*", help="Filenames: vasprun.xml or OUTCAR and POSCAR"
    )
    args = parser.parse_args()
    return args


def run():
    """Rurn phonopy-vasp-born."""
    args = get_options()

    if args.filenames:
        outcar_filename = args.filenames[0]
    else:
        if args.read_outcar:
            outcar_filename = "OUTCAR"
        else:
            outcar_filename = "vasprun.xml"

    if len(args.filenames) > 1:
        poscar_filename = args.filenames[1]
    else:
        poscar_filename = "POSCAR"

    if args.primitive_axes:
        vals = [fracval(x) for x in args.primitive_axes.split()]
        if len(vals) == 9:
            primitive_axes = np.array(vals).reshape(3, 3)
        else:
            print("Primitive axes are incorrectly set.")
            sys.exit(1)
    else:
        primitive_axes = np.eye(3)

    if args.supercell_matrix:
        vals = [int(x) for x in args.supercell_matrix.split()]
        if len(vals) == 9:
            supercell_matrix = np.reshape(np.array(vals, dtype="intc"), (3, 3))
        elif len(vals) == 3:
            supercell_matrix = np.diag(np.array(vals, dtype="intc"))
        else:
            print("Supercell matrix is incorrectly set.")
            sys.exit(1)
    else:
        supercell_matrix = np.eye(3, dtype="intc")

    with warnings.catch_warnings():
        # To catch warnings as error, set warnings.simplefilter("error")
        warnings.simplefilter("always")

        try:
            if args.read_outcar:
                borns, epsilon, atom_indices = get_born_OUTCAR(
                    poscar_filename=poscar_filename,
                    outcar_filename=outcar_filename,
                    primitive_matrix=primitive_axes,
                    supercell_matrix=supercell_matrix,
                    symmetrize_tensors=args.symmetrize_tensors,
                    symprec=args.symprec,
                )
            else:
                borns, epsilon, atom_indices = get_born_vasprunxml(
                    filename=outcar_filename,
                    primitive_matrix=primitive_axes,
                    supercell_matrix=supercell_matrix,
                    symmetrize_tensors=args.symmetrize_tensors,
                    symprec=args.symprec,
                )
        except UserWarning:
            text = "# Symmetry broken"
            lines = [text]
            print("\n".join(lines))
            sys.exit(0)

        try:
            text = "# epsilon and Z* of atoms "
            text += " ".join(["%d" % n for n in atom_indices + 1])
            lines = [text]
            lines.append(("%13.8f " * 9) % tuple(epsilon.flatten()))
            for z in borns:
                lines.append(("%13.8f " * 9) % tuple(z.flatten()))
        except Exception:
            sys.exit(1)
        else:
            print("\n".join(lines))
            sys.exit(0)

    sys.exit(1)



"""SIESTA calculator interface."""

# Copyright (C) 2015 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re
import sys

import numpy as np

from phonopy.file_IO import iter_collect_forces
from phonopy.interface.vasp import check_forces, get_drift_forces
from phonopy.physical_units import get_physical_units
from phonopy.structure.atoms import PhonopyAtoms


def parse_set_of_forces(num_atoms, forces_filenames, verbose=True):
    """Parse forces from output files."""
    hook = ""  # Just for skipping the first line
    is_parsed = True
    force_sets = []
    for i, filename in enumerate(forces_filenames):
        if verbose:
            sys.stdout.write("%d. " % (i + 1))
        siesta_forces = iter_collect_forces(
            filename, num_atoms, hook, [1, 2, 3], word=""
        )
        if check_forces(siesta_forces, num_atoms, filename, verbose=verbose):
            drift_force = get_drift_forces(
                siesta_forces, filename=filename, verbose=verbose
            )
            force_sets.append(np.array(siesta_forces) - drift_force)
        else:
            is_parsed = False

    if is_parsed:
        return force_sets
    else:
        return []


def read_siesta(filename):
    """Read crystal structure."""
    siesta_in = SiestaIn(open(filename).read())
    numbers = siesta_in._tags["atomicnumbers"]
    alat = siesta_in._tags["latticeconstant"]
    lattice = siesta_in._tags["latticevectors"]
    positions = siesta_in._tags["atomiccoordinates"]
    atypes = siesta_in._tags["chemicalspecieslabel"]
    cell = PhonopyAtoms(numbers=numbers, cell=lattice, scaled_positions=positions)

    coordformat = siesta_in._tags["atomiccoordinatesformat"]
    if coordformat == "fractional" or coordformat == "scaledbylatticevectors":
        cell.set_scaled_positions(positions)
    elif coordformat == "scaledcartesian":
        cell.set_positions(np.array(positions) * alat)
    elif coordformat == "notscaledcartesianang" or coordformat == "ang":
        cell.set_positions(np.array(positions) / get_physical_units().Bohr)
    elif coordformat == "notscaledcartesianbohr" or coordformat == "bohr":
        cell.set_positions(np.array(positions))
    else:
        print(
            "The format %s for the AtomicCoordinatesFormat is not "
            "implemented." % coordformat
        )
        sys.exit(1)

    return cell, atypes


def write_siesta(filename, cell, atypes):
    """Write cell to file."""
    with open(filename, "w") as w:
        w.write(get_siesta_structure(cell, atypes))


def write_supercells_with_displacements(
    supercell, cells_with_displacements, ids, atypes, pre_filename="supercell", width=3
):
    """Write supercells with displacements to files."""
    write_siesta("%s.fdf" % pre_filename, supercell, atypes)
    for i, cell in zip(ids, cells_with_displacements):
        filename = "{pre_filename}-{0:0{width}}.fdf".format(
            i, pre_filename=pre_filename, width=width
        )
        write_siesta(filename, cell, atypes)


def get_siesta_structure(cell, atypes):
    """Return SIESTA structure in text."""
    lattice = cell.get_cell()
    positions = cell.get_scaled_positions()
    chemical_symbols = cell.get_chemical_symbols()

    lines = ""

    lines += "NumberOfAtoms %d\n\n" % len(positions)

    lines += "%block LatticeVectors\n"
    lines += ((" %21.16f" * 3 + "\n") * 3) % tuple(lattice.ravel())
    lines += "%endblock LatticeVectors\n\n"

    lines += "AtomicCoordinatesFormat  Fractional\n\n"

    lines += "LatticeConstant 1.0 Bohr\n\n"

    lines += "%block AtomicCoordinatesAndAtomicSpecies\n"
    for pos, i in zip(positions, chemical_symbols):
        lines += ("%21.16lf" * 3 + " %d\n") % tuple(pos.tolist() + [atypes[i]])
    lines += "%endblock AtomicCoordinatesAndAtomicSpecies\n"

    return lines


class SiestaIn:
    """Class to create SIESTA input file."""

    _num_regex = r"([+-]?\d+(?:\.\d*)?(?:[eE][-+]?\d+)?)"
    _tags = {
        "latticeconstant": 1.0,
        "latticeconstantunit": None,
        "chemicalspecieslabel": None,
        "atomiccoordinatesformat": None,
        "atomicnumbers": None,
        "atomicspecies": None,
        "atomiccoordinates": None,
    }

    def __init__(self, lines):
        """Init method."""
        self._collect(lines)

    def _collect(self, lines):
        """Collect values.

        This routine reads the following from the Siesta file:
        - atomic positions
        - cell_parameters
        - atomic_species

        """
        for tag, value, unit in re.findall(
            r"([\.A-Za-z]+)\s+%s\s+([A-Za-z]+)?" % self._num_regex, lines
        ):
            tag = tag.lower()
            unit = unit.lower()
            if tag == "latticeconstant":
                self._tags["latticeconstantunit"] = unit.capitalize()
                if unit == "ang":
                    self._tags[tag] = float(value) / get_physical_units().Bohr
                elif unit == "bohr":
                    self._tags[tag] = float(value)
                else:
                    raise ValueError("Unknown LatticeConstant unit: {}".format(unit))

        for tag, value in re.findall(r"([\.A-Za-z]+)[ \t]+([a-zA-Z]+)", lines):
            tag = tag.replace("_", "").lower()
            if tag == "atomiccoordinatesformat":
                self._tags[tag] = value.strip().lower()

        # check if the necessary tags are present
        self._check_present("atomiccoordinatesformat")
        acell = self._tags["latticeconstant"]

        # capture the blocks
        blocks = re.findall(
            r"%block\s+([A-Za-z_]+)\s*\n((?:.+\n)+?(?=(?:\s+)?%endblock))",
            lines,
            re.MULTILINE,
        )
        for tag, block in blocks:
            tag = tag.replace("_", "").lower()
            if tag == "chemicalspecieslabel":
                block_array = block.split("\n")[:-1]
                self._tags["atomicnumbers"] = dict(
                    [map(int, species.split()[:2]) for species in block_array]
                )
                self._tags[tag] = dict(
                    [
                        (lambda x: (x[2], int(x[0])))(species.split())
                        for species in block_array
                    ]
                )
            elif tag == "latticevectors":
                self._tags[tag] = [
                    [float(v) * acell for v in vector.split()]
                    for vector in block.split("\n")[:3]
                ]
            elif tag == "atomiccoordinatesandatomicspecies":
                block_array = block.split("\n")[:-1]
                self._tags["atomiccoordinates"] = [
                    [float(x) for x in atom.split()[:3]] for atom in block_array
                ]
                self._tags["atomicspecies"] = [
                    int(atom.split()[3]) for atom in block_array
                ]

        # check if the block are present
        self._check_present("atomicspecies")
        self._check_present("atomiccoordinates")
        self._check_present("latticevectors")
        self._check_present("chemicalspecieslabel")

        # translate the atomicspecies to atomic numbers
        self._tags["atomicnumbers"] = [
            self._tags["atomicnumbers"][atype] for atype in self._tags["atomicspecies"]
        ]

    def _check_present(self, tag):
        if not self._tags[tag]:
            print("%s not present" % tag)
            sys.exit(1)

    def __str__(self):
        """Return tags."""
        return self._tags


if __name__ == "__main__":
    from phonopy.structure.symmetry import Symmetry

    cell, atypes = read_siesta(sys.argv[1])
    symmetry = Symmetry(cell)
    print("# %s" % symmetry.get_international_table())
    print(get_siesta_structure(cell, atypes))



"""Pytest configuration."""

from __future__ import annotations

import pathlib
from collections.abc import Callable

import numpy as np
import pytest

import phonopy
from phonopy import Phonopy
from phonopy.structure.atoms import PhonopyAtoms

cwd = pathlib.Path(__file__).parent


@pytest.fixture(scope="session")
def ph_si() -> Phonopy:
    """Return Phonopy class instance of Si-prim 2x2x2."""
    yaml_filename = cwd / "phonopy_params_Si.yaml"
    return phonopy.load(
        yaml_filename,
        is_compact_fc=False,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_nacl() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2."""
    yaml_filename = cwd / "phonopy_disp_NaCl.yaml"
    force_sets_filename = cwd / "FORCE_SETS_NaCl"
    born_filename = cwd / "BORN_NaCl"
    return phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        born_filename=born_filename,
        is_compact_fc=False,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_nacl_wang() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2."""
    yaml_filename = cwd / "phonopy_disp_NaCl.yaml"
    force_sets_filename = cwd / "FORCE_SETS_NaCl"
    born_filename = cwd / "BORN_NaCl"
    ph = phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        born_filename=born_filename,
        is_compact_fc=False,
        log_level=1,
        produce_fc=True,
    )
    nac_params = ph.nac_params
    nac_params["method"] = "wang"
    ph.nac_params = nac_params
    return ph


@pytest.fixture(scope="session")
def ph_nacl_nofcsym() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2 without symmetrizing fc2."""
    yaml_filename = cwd / "phonopy_disp_NaCl.yaml"
    force_sets_filename = cwd / "FORCE_SETS_NaCl"
    born_filename = cwd / "BORN_NaCl"
    return phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        born_filename=born_filename,
        symmetrize_fc=False,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_nacl_compact_fcsym() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2 with compact fc2."""
    yaml_filename = cwd / "phonopy_disp_NaCl.yaml"
    force_sets_filename = cwd / "FORCE_SETS_NaCl"
    born_filename = cwd / "BORN_NaCl"
    return phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        born_filename=born_filename,
        is_compact_fc=True,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_nacl_nonac() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2 without NAC."""
    yaml_filename = cwd / "phonopy_disp_NaCl.yaml"
    force_sets_filename = cwd / "FORCE_SETS_NaCl"
    return phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        is_nac=False,
        is_compact_fc=False,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_nacl_nonac_compact_fc() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2 without NAC with compact fc2."""
    yaml_filename = cwd / "phonopy_disp_NaCl.yaml"
    force_sets_filename = cwd / "FORCE_SETS_NaCl"
    return phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        is_nac=False,
        is_compact_fc=True,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_nacl_nonac_dense_svecs() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2 without NAC with dense svecs."""
    yaml_filename = cwd / "phonopy_disp_NaCl.yaml"
    force_sets_filename = cwd / "FORCE_SETS_NaCl"
    return phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        is_nac=False,
        is_compact_fc=True,
        store_dense_svecs=True,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_nacl_rd() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2 with RD results.

    This data contains supercell energies and NAC params.

    """
    yaml_filename = cwd / "phonopy_params_NaCl-rd.yaml.xz"
    return phonopy.load(yaml_filename, log_level=1, produce_fc=False)


@pytest.fixture(scope="session")
def ph_nacl_rd_symfc() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2 with RD results.

    This data contains supercell energies and NAC params. Symfc is used to
    compute force constants.

    """
    pytest.importorskip("symfc")
    pytest.importorskip("spglib", minversion="2.5")

    yaml_filename = cwd / "phonopy_params_NaCl-rd.yaml.xz"
    return phonopy.load(
        yaml_filename, log_level=1, fc_calculator="symfc", is_compact_fc=True
    )


@pytest.fixture(scope="session")
def ph_nacl_fd() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2 with RD results.

    This data contains supercell energies and NAC params.

    """
    yaml_filename = cwd / "phonopy_params_NaCl-fd.yaml.xz"
    return phonopy.load(yaml_filename, log_level=1, produce_fc=False)


@pytest.fixture(scope="session")
def ph_sno2() -> Phonopy:
    """Return Phonopy class instance of rutile SnO2 2x2x3."""
    yaml_filename = cwd / "phonopy_disp_SnO2.yaml"
    force_sets_filename = cwd / "FORCE_SETS_SnO2"
    born_filename = cwd / "BORN_SnO2"
    return phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        born_filename=born_filename,
        is_compact_fc=False,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_tio2() -> Phonopy:
    """Return Phonopy class instance of anataze TiO2 3x3x1."""
    yaml_filename = cwd / "phonopy_disp_TiO2.yaml"
    force_sets_filename = cwd / "FORCE_SETS_TiO2"
    born_filename = cwd / "BORN_TiO2"
    return phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        born_filename=born_filename,
        is_compact_fc=False,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_zr3n4() -> Phonopy:
    """Return Phonopy class instance of anataze Zr3N4 1x1x1."""
    yaml_filename = cwd / "phonopy_params_Zr3N4.yaml"
    return phonopy.load(
        yaml_filename, is_compact_fc=False, log_level=1, produce_fc=True
    )


@pytest.fixture(scope="session")
def ph_tipn3() -> Phonopy:
    """Return Phonopy class instance of anataze TiPN3 4x2x1."""
    yaml_filename = cwd / "phonopy_params_TiPN3.yaml.xz"
    return phonopy.load(
        yaml_filename, is_compact_fc=False, log_level=1, produce_fc=True
    )


@pytest.fixture(scope="session")
def ph_srtio3() -> Phonopy:
    """Return Phonopy class instance of anataze SrTiO3 3x3x3."""
    yaml_filename = cwd / "phonopy_SrTiO3.yaml.xz"
    return phonopy.load(
        yaml_filename, is_compact_fc=False, log_level=1, produce_fc=True
    )


@pytest.fixture(scope="session")
def ph_kcl() -> Phonopy:
    """Return Phonopy class instance of KCl for mlpsscha_KCl-120.pmlp."""
    yaml_filename = cwd / "phonopy_KCl.yaml"
    return phonopy.load(yaml_filename, produce_fc=False, log_level=1)


@pytest.fixture(scope="session")
def ph_nacl_gruneisen() -> tuple[Phonopy, Phonopy, Phonopy]:
    """Return Phonopy class instances of NaCl 2x2x2 at three volumes."""
    ph0 = phonopy.load(
        cwd / "phonopy_params_NaCl-1.00.yaml.xz",
        log_level=1,
        produce_fc=True,
    )
    ph_minus = phonopy.load(
        cwd / "phonopy_params_NaCl-0.995.yaml.xz",
        log_level=1,
        produce_fc=True,
    )
    ph_plus = phonopy.load(
        cwd / "phonopy_params_NaCl-1.005.yaml.xz",
        log_level=1,
        produce_fc=True,
    )
    return ph0, ph_minus, ph_plus


@pytest.fixture(scope="session")
def convcell_sio2() -> PhonopyAtoms:
    """Return PhonopyAtoms class instance of rutile SiO2."""
    symbols = ["Si"] * 2 + ["O"] * 4
    lattice = [[4.65, 0, 0], [0, 4.75, 0], [0, 0, 3.25]]
    points = [
        [0.0, 0.0, 0.0],
        [0.5, 0.5, 0.5],
        [0.3, 0.3, 0.0],
        [0.7, 0.7, 0.0],
        [0.2, 0.8, 0.5],
        [0.8, 0.2, 0.5],
    ]
    return PhonopyAtoms(cell=lattice, scaled_positions=points, symbols=symbols)


@pytest.fixture(scope="session")
def convcell_al2o3() -> PhonopyAtoms:
    """Return PhonopyAtoms class instance of rutile Al2O3."""
    symbols = ["Al"] * 12 + ["O"] * 18
    lattice = [
        [4.7742263452989944, 0.0000000000000000, 0.0000000000000000],
        [-2.3871131726494972, 4.1346012984458653, 0.0000000000000000],
        [-0.0000000000000000, 0.0000000000000000, 13.0113593913272219],
    ]
    points = [
        [0.3333333333333357, 0.6666666666666643, 0.0187603745624614],
        [0.3333333333333357, 0.6666666666666643, 0.3145729587708743],
        [0.0000000000000000, 0.0000000000000000, 0.1479062921042101],
        [0.6666666666666643, 0.3333333333333357, 0.1854270412291256],
        [0.0000000000000000, 0.0000000000000000, 0.3520937078957900],
        [0.0000000000000000, 0.0000000000000000, 0.6479062921042100],
        [0.6666666666666643, 0.3333333333333357, 0.4812396254375386],
        [0.3333333333333357, 0.6666666666666643, 0.5187603745624614],
        [0.6666666666666643, 0.3333333333333357, 0.6854270412291257],
        [0.6666666666666643, 0.3333333333333357, 0.9812396254375386],
        [0.3333333333333357, 0.6666666666666643, 0.8145729587708743],
        [0.0000000000000000, 0.0000000000000000, 0.8520937078957900],
        [0.3058086537880128, 0.0000000000000000, 0.2500000000000000],
        [0.3608580128786515, 0.3333333333333357, 0.0833333333333357],
        [0.0000000000000000, 0.3058086537880128, 0.2500000000000000],
        [0.6666666666666643, 0.0275246795453229, 0.0833333333333357],
        [0.6941913462119872, 0.6941913462119872, 0.2500000000000000],
        [0.9724753204546771, 0.6391419871213485, 0.0833333333333357],
        [0.9724753204546771, 0.3333333333333357, 0.5833333333333357],
        [0.0275246795453229, 0.6666666666666643, 0.4166666666666643],
        [0.6666666666666643, 0.6391419871213485, 0.5833333333333357],
        [0.3333333333333357, 0.3608580128786515, 0.4166666666666643],
        [0.3608580128786515, 0.0275246795453229, 0.5833333333333357],
        [0.6391419871213485, 0.9724753204546771, 0.4166666666666643],
        [0.6391419871213485, 0.6666666666666643, 0.9166666666666643],
        [0.6941913462119872, 0.0000000000000000, 0.7500000000000000],
        [0.3333333333333357, 0.9724753204546771, 0.9166666666666643],
        [0.0000000000000000, 0.6941913462119872, 0.7500000000000000],
        [0.0275246795453229, 0.3608580128786515, 0.9166666666666643],
        [0.3058086537880128, 0.3058086537880128, 0.7500000000000000],
    ]
    return PhonopyAtoms(cell=lattice, scaled_positions=points, symbols=symbols)


@pytest.fixture(scope="session")
def primcell_si() -> PhonopyAtoms:
    """Return PhonopyAtoms class instance of primitive cell of Si."""
    symbols = ["Si"] * 2
    lattice = [[0, 2.73, 2.73], [2.73, 0, 2.73], [2.73, 2.73, 0]]
    points = [[0.75, 0.75, 0.75], [0.5, 0.5, 0.5]]
    return PhonopyAtoms(cell=lattice, scaled_positions=points, symbols=symbols)


@pytest.fixture(scope="session")
def nacl_unitcell_order1() -> PhonopyAtoms:
    """Return Phonopy class instance of only NaCl unitcell with order-1."""
    yaml_filename = cwd / "phonopy_NaCl_unitcell1.yaml"
    return phonopy.load(yaml_filename, log_level=1, produce_fc=False).unitcell


@pytest.fixture(scope="session")
def nacl_unitcell_order2() -> PhonopyAtoms:
    """Return Phonopy class instance of only NaCl unitcell with order-2."""
    yaml_filename = cwd / "phonopy_NaCl_unitcell2.yaml"
    return phonopy.load(yaml_filename, log_level=1, produce_fc=False).unitcell


@pytest.fixture(scope="session")
def primcell_nacl() -> PhonopyAtoms:
    """Return PhonopyAtoms class instance of primitive cell of NaCl."""
    symbols = ["Na", "Cl"]
    x = 5.6903014761756712 / 2
    lattice = [[0, x, x], [x, 0, x], [x, x, 0]]
    points = [[0, 0, 0], [0.5, 0.5, 0.5]]
    return PhonopyAtoms(cell=lattice, scaled_positions=points, symbols=symbols)


@pytest.fixture(scope="session")
def convcell_cr() -> PhonopyAtoms:
    """Return PhonopyAtoms class instance of primitive cell of Cr."""
    symbols = ["Cr"] * 2
    a = 2.812696943681890
    lattice = [[a, 0, 0], [0, a, 0], [0, 0, a]]
    points = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]]
    return PhonopyAtoms(cell=lattice, scaled_positions=points, symbols=symbols)


@pytest.fixture(scope="session")
def helper_methods() -> Callable:
    """Return methods to compare cells."""

    class HelperMethods:
        @classmethod
        def compare_cells_with_order(
            cls, cell: PhonopyAtoms, cell_ref: PhonopyAtoms, symprec=1e-5
        ) -> None:
            """Compare two cells with the same orders of positions."""
            np.testing.assert_allclose(cell.cell, cell_ref.cell, atol=symprec)
            cls.compare_positions_with_order(
                cell.scaled_positions, cell_ref.scaled_positions, cell.cell
            )
            np.testing.assert_array_equal(cell.numbers, cell_ref.numbers)
            np.testing.assert_allclose(cell.masses, cell_ref.masses, atol=symprec)
            if cell.magnetic_moments is None:
                assert cell_ref.magnetic_moments is None
            else:
                np.testing.assert_allclose(
                    cell.magnetic_moments, cell_ref.magnetic_moments, atol=symprec
                )

        @classmethod
        def compare_positions_with_order(
            cls, pos, pos_ref, lattice, symprec=1e-5
        ) -> None:
            """Compare two lists of positions and orders.

            lattice :
                Basis vectors in row vectors.

            """
            diff = pos - pos_ref
            diff -= np.rint(diff)
            dist = (np.dot(diff, lattice) ** 2).sum(axis=1)
            assert (dist < symprec).all()

        @classmethod
        def compare_cells(
            cls, cell: PhonopyAtoms, cell_ref: PhonopyAtoms, symprec=1e-5
        ) -> None:
            """Compare two cells where position orders can be different."""
            np.testing.assert_allclose(cell.cell, cell_ref.cell, atol=symprec)

            indices = cls.compare_positions_in_arbitrary_order(
                cell.scaled_positions, cell_ref.scaled_positions, cell.cell
            )
            np.testing.assert_array_equal(cell.numbers, cell_ref.numbers[indices])
            np.testing.assert_allclose(
                cell.masses, cell_ref.masses[indices], atol=symprec
            )
            if cell.magnetic_moments is None:
                assert cell_ref.magnetic_moments is None
            else:
                np.testing.assert_allclose(
                    cell.magnetic_moments,
                    cell_ref.magnetic_moments[indices],
                    atol=symprec,
                )

        @classmethod
        def compare_positions_in_arbitrary_order(
            cls, pos_in, pos_ref, lattice, symprec=1e-5
        ) -> list:
            """Compare two sets of positions irrespective of orders.

            lattice :
                Basis vectors in row vectors.

            """
            indices = []
            for pos in pos_in:
                diff = pos_ref - pos
                diff -= np.rint(diff)
                dist = (np.dot(diff, lattice) ** 2).sum(axis=1)
                matches = np.where(dist < symprec)[0]
                assert len(matches) == 1
                indices.append(matches[0])
            return indices

    return HelperMethods



"""Calculation of Smith normal form of 3x3 matrix."""

# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np


#
# Smith normal form for 3x3 integer matrix
# This code is maintained at https://github.com/atztogo/snf3x3.
#
class SNF3x3:
    """Smith normal form for 3x3 matrix.

    Input 3x3 interger matrix A is transformed to 3x3 diagonal interger
    matrix (D) by two 3x3 interger matrices P and Q, which is written as

        D = PAQ

    with abs(det(A)) = det(D), det(P) = 1, det(Q) = sgn(det(A)).

    The algorithm is implemented following
    https://en.wikipedia.org/wiki/Smith_normal_form,
    but the diagonal elements don't follow the rule of it.

    Usage
    -----
    snf = SNF3x3(int_mat)
    snf.run()

    Attributes
    ----------
    D, P, Q : ndarray
        These 3x3 interger matrices explained above.
        shape=(3, 3), dtype='int64'

    """

    def __init__(self, A):
        """Init method.

        Parameters
        ----------
        A : array_like
            3x3 interger matrix.
            shape=(3, 3)

        """
        self._A_orig = np.array(A, dtype="int64", order="C")
        self._A = np.array(A, dtype="int64", order="C")
        self._Ps = []
        self._Qs = []
        self._L = []
        self._P = None
        self._Q = None
        self._D = None
        self._attempt = 0

    def run(self):
        """Calculate SNF."""
        for _ in self:
            pass

        A = np.dot(self._P, np.dot(self._A_orig, self._Q))
        assert (A == self._A).all()

    def __iter__(self):
        """Define iterator over calculation steps."""
        return self

    def __next__(self):
        """Calculate an algorithm step."""
        self._attempt += 1
        if self._first():
            if self._second():
                self._finalize()
                raise StopIteration
        return self._attempt

    @property
    def A(self):
        """Return A of D = PAQ."""
        return self._A

    @property
    def D(self):
        """Return D of D = PAQ."""
        return self._D

    @property
    def P(self):
        """Return P of D = PAQ."""
        return self._P

    @property
    def Q(self):
        """Return Q of D = PAQ."""
        return self._Q

    def _first(self):
        self._first_one_loop()
        A = self._A
        if A[1, 0] == 0 and A[2, 0] == 0:
            return True
        elif A[1, 0] % A[0, 0] == 0 and A[2, 0] % A[0, 0] == 0:
            self._first_finalize()
            self._Ps += self._L
            self._L = []
            return True
        else:
            return False

    def _first_one_loop(self):
        self._first_column()
        self._Ps += self._L
        self._L = []
        self._A[:] = self._A.T
        self._first_column()
        self._Qs += self._L
        self._L = []
        self._A[:] = self._A.T

    def _first_column(self):
        i = self._search_first_pivot()
        if i > 0:
            self._swap_rows(0, i)
        if i < 0:
            raise RuntimeError("Determinant is 0.")

        if self._A[1, 0] != 0:
            self._zero_first_column(1)
        if self._A[2, 0] != 0:
            self._zero_first_column(2)

    def _zero_first_column(self, j):
        A = self._A
        r, s, t = xgcd([A[0, 0], A[j, 0]])
        self._set_zero(0, j, A[0, 0], A[j, 0], r, s, t)

    def _search_first_pivot(self):
        for i in range(3):  # column index
            if self._A[i, 0] != 0:
                return i
        return -1

    def _first_finalize(self):
        """Set zeros along the first colomn except for A[0, 0].

        This is possible only when A[1,0] and A[2,0] are dividable by A[0,0].

        """
        A = self._A
        L = np.eye(3, dtype="int64")
        L[1, 0] = -A[1, 0] // A[0, 0]
        L[2, 0] = -A[2, 0] // A[0, 0]
        self._L.append(L.copy())
        self._A[:] = np.dot(L, self._A)

    def _second(self):
        """Find Smith normal form for Right-low 2x2 matrix."""
        self._second_one_loop()
        A = self._A
        if A[2, 1] == 0:
            return True
        elif A[2, 1] % A[1, 1] == 0:
            self._second_finalize()
            self._Ps += self._L
            self._L = []
            return True
        else:
            return False

    def _second_one_loop(self):
        self._second_column()
        self._Ps += self._L
        self._L = []
        self._A[:] = self._A.T
        self._second_column()
        self._Qs += self._L
        self._L = []
        self._A[:] = self._A.T

    def _second_column(self):
        """Right-low 2x2 matrix.

        Assume elements in first row and column are all zero except for A[0,0].

        """
        if self._A[1, 1] == 0 and self._A[2, 1] != 0:
            self._swap_rows(1, 2)

        if self._A[2, 1] != 0:
            self._zero_second_column()

    def _zero_second_column(self):
        A = self._A
        r, s, t = xgcd([A[1, 1], A[2, 1]])
        self._set_zero(1, 2, A[1, 1], A[2, 1], r, s, t)

    def _second_finalize(self):
        """Set zero at A[2, 1].

        This is possible only when A[2,1] is dividable by A[1,1].

        """
        A = self._A
        L = np.eye(3, dtype="int64")
        L[2, 1] = -A[2, 1] // A[1, 1]
        self._L.append(L.copy())
        self._A[:] = np.dot(L, self._A)

    def _finalize(self):
        for i in range(3):
            if self._A[i, i] < 0:
                self._flip_sign_row(i)
                self._Ps += self._L
                self._L = []
        self._finalize_sort()
        self._finalize_disturb(0, 1)
        self._first()
        self._finalize_sort()
        self._finalize_disturb(1, 2)
        self._second()
        self._set_PQ()

    def _finalize_sort(self):
        if self._A[0, 0] > self._A[1, 1]:
            self._swap_diag_elems(0, 1)
        if self._A[1, 1] > self._A[2, 2]:
            self._swap_diag_elems(1, 2)
        if self._A[0, 0] > self._A[1, 1]:
            self._swap_diag_elems(0, 1)

    def _finalize_disturb(self, i, j):
        if self._A[j, j] % self._A[i, i] != 0:
            self._A[:] = self._A.T
            self._disturb_rows(i, j)
            self._Qs += self._L
            self._L = []
            self._A[:] = self._A.T

    def _disturb_rows(self, i, j):
        L = np.eye(3, dtype="int64")
        L[i, i] = 1
        L[i, j] = 1
        L[j, i] = 0
        L[j, j] = 1
        self._L.append(L.copy())
        self._A[:] = np.dot(L, self._A)

    def _swap_diag_elems(self, i, j):
        self._swap_rows(i, j)
        self._Ps += self._L
        self._L = []
        self._A[:] = self._A.T
        self._swap_rows(i, j)
        self._Qs += self._L
        self._L = []
        self._A[:] = self._A.T

    def _swap_rows(self, i, j):
        """Swap i and j rows.

        As the side effect, determinant flips.

        """
        L = np.eye(3, dtype="int64")
        L[i, i] = 0
        L[j, j] = 0
        L[i, j] = 1
        L[j, i] = 1
        self._L.append(L.copy())
        self._A[:] = np.dot(L, self._A)

    def _flip_sign_row(self, i):
        """Multiply -1 for all elements in row."""
        L = np.eye(3, dtype="int64")
        L[i, i] = -1
        self._L.append(L.copy())
        self._A[:] = np.dot(L, self._A)

    def _set_zero(self, i, j, a, b, r, s, t):
        """Let A[i, j] be zero based on Bezout's identity.

        [ii ij]
        [ji jj] is a (k,k) minor of original 3x3 matrix.

        """
        L = np.eye(3, dtype="int64")
        L[i, i] = s
        L[i, j] = t
        L[j, i] = -b // r
        L[j, j] = a // r
        self._L.append(L.copy())
        self._A[:] = np.dot(L, self._A)

    def _set_PQ(self):
        P = np.eye(3, dtype="int64")
        for _P in self._Ps:
            P = np.dot(_P, P)
        Q = np.eye(3, dtype="int64")
        for _Q in self._Qs:
            Q = np.dot(Q, _Q.T)

        if self._det(P) < 0:
            P *= -1
            Q *= -1

        self._P = P
        self._Q = Q
        self._D = self._A.copy()

    def _det(self, m):
        return (
            m[0, 0] * (m[1, 1] * m[2, 2] - m[1, 2] * m[2, 1])
            + m[0, 1] * (m[1, 2] * m[2, 0] - m[1, 0] * m[2, 2])
            + m[0, 2] * (m[1, 0] * m[2, 1] - m[1, 1] * m[2, 0])
        )


def xgcd(vals):
    """Calculate extended greatest commond divisor."""
    _xgcd = Xgcd(vals)
    return _xgcd.run()


class Xgcd:
    """Class to implement extended Euclidean algorithm."""

    def __init__(self, vals):
        """Init method."""
        self._vals = np.array(vals, dtype="intc")

    def run(self):
        """Calculate extended GCD."""
        r0, r1 = self._vals
        s0 = 1
        s1 = 0
        t0 = 0
        t1 = 1
        for _ in range(1000):
            r0, r1, s0, s1, t0, t1 = self._step(r0, r1, s0, s1, t0, t1)
            if r1 == 0:
                break

        assert r0 == self._vals[0] * s0 + self._vals[1] * t0

        self._rst = np.array([r0, s0, t0], dtype="intc")

        return self._rst

    def _step(self, r0, r1, s0, s1, t0, t1):
        q, m = divmod(r0, r1)
        if m < 0:
            if r1 > 0:
                m += r1
                q -= 1
            if r1 < 0:
                m -= r1
                q += 1
        r2 = m
        s2 = s0 - q * s1
        t2 = t0 - q * t1
        return r1, r2, s1, s2, t1, t2

    def __str__(self):
        """Show GCD relation."""
        v = self._vals
        r, s, t = self._rst
        return "%d = %d * (%d) + %d * (%d)" % (r, v[0], s, v[1], t)



"""Tests of custom exceptions."""

import pytest

from phonopy import Phonopy
from phonopy.exception import ForcesetsNotFoundError


def test_ForcesetsNotFoundError(nacl_unitcell_order1: Phonopy):
    """Test of ForcesetsNotFoundError."""
    ph = Phonopy(nacl_unitcell_order1)
    ph.generate_displacements()
    with pytest.raises(ForcesetsNotFoundError):
        ph.produce_force_constants()



# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import lzma
import os
import sys

import h5py
import numpy as np

from phonopy.phonon.band_structure import BandPlot

try:
    import yaml
except ImportError:
    print("You need to install python-yaml.")
    sys.exit(1)

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


def _get_label_for_latex(label):
    return label.replace("_", r"\_")


def _get_max_frequency(frequencies):
    return max([np.max(fq) for fq in frequencies])


def _find_wrong_path_connections(all_path_connections):
    for i, path_connections in enumerate(all_path_connections):
        if path_connections != all_path_connections[0]:
            return i
    return 0


def _arrange_band_data(distances, frequencies, qpoints, segment_nqpoints, label_pairs):
    i = 0
    freq_list = []
    dist_list = []
    qpt_list = []
    for nq in segment_nqpoints:
        freq_list.append(frequencies[i : (i + nq)])
        dist_list.append(distances[i : (i + nq)])
        qpt_list.append(qpoints[i : (i + nq)])
        i += nq

    if not label_pairs:
        labels = None
        path_connections = []
        if len(qpt_list) > 1:
            for i, qpts in enumerate(qpt_list[1:]):
                if (np.abs(qpt_list[i][-1] - qpts[0]) < 1e-5).all():
                    path_connections.append(True)
                else:
                    path_connections.append(False)
        path_connections += [
            False,
        ]
    else:
        labels = []
        path_connections = []
        if len(label_pairs) > 1:
            for i, pair in enumerate(label_pairs[1:]):
                labels.append(label_pairs[i][0])
                if label_pairs[i][1] != pair[0]:
                    labels.append(label_pairs[i][1])
                    path_connections.append(False)
                else:
                    path_connections.append(True)
            if label_pairs[-2][1] != label_pairs[-1][1]:
                labels += label_pairs[-1]
            else:
                labels.append(label_pairs[-1][1])
        else:
            labels += label_pairs[0]
        path_connections += [
            False,
        ]

    return labels, path_connections, freq_list, dist_list


def _savefig(plt, file, fonttype=42, family="serif"):
    plt.rcParams["pdf.fonttype"] = fonttype
    plt.rcParams["font.family"] = family
    plt.savefig(file)


def _get_dos(d, f, dmax):
    pdos = []
    for f1, f2, d1, d2 in zip(f[:-1], f[1:], d[:-1], d[1:]):
        pdos += _cut_dos([f1, d1], [f2, d2], dmax)
    return np.transpose(pdos)


def _cut_dos(p1: list[float, float], p2: list[float, float], dmax: float):
    """Cut DOS at dmax.

    ** Cutting by fmin and fmax should be implemented someday. **

    Parameters
    ----------
    p1 : list[float, float]
        (f1, d1)
    p2 : list[float, float]
        (f2, d2)

    When d1 < dmax and d2 < dmax
        p1: (f1, d1)
        p2: (f2, d2)
    When d1 < dmax and d2 > dmax
        p1: (f1, d1)
        pi: (fi, dmax), f1 < fi < f2
        p2: (f2, dmax)
    Whend1 > dmax and d2 < dmax,
        p1: (f1, dmax)
        pi: (fi, dmax), f1 < fi < f2
        p2: (f2, d2)
    When d2 > dmax and d1 > dmax
        p1: (f1, dmax)
        p2: (f2, dmax)

    """

    def _get_fi(p1, p2, dmax):
        df = p2[0] - p1[0]
        dd = p2[1] - p1[1]
        fi = (dmax - p1[1]) / dd * df + p1[0]
        return fi

    if p1[1] < dmax and p2[1] < dmax:
        return [p1, p2]
    elif p1[1] < dmax and p2[1] > dmax:
        fi = _get_fi(p1, p2, dmax)
        return [p1, [fi, dmax], [p2[0], dmax]]
    elif p1[1] > dmax and p2[1] < dmax:
        fi = _get_fi(p1, p2, dmax)
        return [[p1[0], dmax], [fi, dmax], p2]
    else:
        return [[p1[0], dmax], [p1[0], dmax]]


def _read_band_yaml(filename):
    _, ext = os.path.splitext(filename)
    if ext == ".xz" or ext == ".lzma":
        with lzma.open(filename) as f:
            data = yaml.load(f, Loader=Loader)
    elif ext == ".gz":
        import gzip

        with gzip.open(filename) as f:
            data = yaml.load(f, Loader=Loader)
    else:
        with open(filename, "r") as f:
            data = yaml.load(f, Loader=Loader)

    frequencies = []
    distances = []
    qpoints = []
    labels = []
    for v in data["phonon"]:
        if "label" in v:
            labels.append(v["label"])
        else:
            labels.append(None)
        frequencies.append([f["frequency"] for f in v["band"]])
        qpoints.append(v["q-position"])
        distances.append(v["distance"])

    if "labels" in data:
        labels = data["labels"]
    elif all(x is None for x in labels):
        labels = []

    return (
        np.array(distances),
        np.array(frequencies),
        np.array(qpoints),
        data["segment_nqpoint"],
        labels,
    )


def _read_band_hdf5(filename):
    with h5py.File(filename, "r") as data:
        f2 = data["frequency"][:]
        d1 = data["distance"][:]
        # lbl = data['label'][:]

        labels_path = []
        for x in data["label"][:]:
            labels_path.append([y.decode("utf-8") for y in x])

        frequencies = f2.reshape((f2.shape[0] * f2.shape[1], f2.shape[2]))
        distances = []
        for x in d1:
            for y in x:
                distances.append(y)
        qpoints = [i for j in data["path"][:] for i in j]

        seg_pt = data["segment_nqpoint"][:]

        # nqpt = data['nqpoint'][0]

    return (
        np.array(distances),
        np.array(frequencies),
        np.array(qpoints),
        seg_pt,
        labels_path,
    )


def _read_dos_dat(filename, pdos_indices=None, dos_factor=None):
    dos = []
    frequencies = []
    for line in open(filename):
        if line.strip()[0] == "#":
            continue
        ary = [float(x) for x in line.split()]
        frequencies.append(ary.pop(0))
        dos.append(ary)
    dos = np.array(dos)
    frequencies = np.array(frequencies)

    if pdos_indices:
        pi = []
        for nums in pdos_indices.split(","):
            pi.append([int(x) - 1 for x in nums.split()])
        dos_sum = []
        for indices in pi:
            dos_sum.append(dos[:, indices].sum(axis=1))
        dos_sum.append(dos.sum(axis=1))
        dos = np.transpose(dos_sum)

    if dos_factor:
        dos *= dos_factor

    return frequencies, dos


def get_options():
    """Parse options."""
    import argparse

    parser = argparse.ArgumentParser(description="Phonopy bandplot command-line-tool")
    parser.set_defaults(
        hdf5=None,
        band_labels=None,
        dos=None,
        dos_max=None,
        dos_min=None,
        dos_factor=None,
        factor=1.0,
        f_max=None,
        f_min=None,
        is_gnuplot=False,
        is_legacy_plot=False,
        is_points=False,
        is_vertical_line=False,
        output_filename=None,
        pdos_indices=None,
        xlabel=None,
        ylabel=None,
        show_legend=False,
        title=None,
    )
    parser.add_argument(
        "--hdf5", dest="is_hdf5", action="store_true", help="Read HDF5 format"
    )
    parser.add_argument(
        "--dmax",
        dest="dos_max",
        type=float,
        help="Maximum DOS plotted (legacy plot only)",
    )
    parser.add_argument(
        "--dmin",
        dest="dos_min",
        type=float,
        help="Minimum DOS plotted (legacy plot only)",
    )
    parser.add_argument(
        "--dos", dest="dos", help="Read dos.dat type file and plot with band structure"
    )
    parser.add_argument(
        "--dos-factor",
        dest="dos_factor",
        type=float,
        help="Factor to be multiplied with DOS (legacy plot only)",
    )
    parser.add_argument(
        "--factor",
        dest="factor",
        type=float,
        help="Conversion factor to favorite frequency unit",
    )
    parser.add_argument(
        "--fmax", dest="f_max", type=float, help="Maximum frequency plotted"
    )
    parser.add_argument(
        "--fmin", dest="f_min", type=float, help="Minimum frequency plotted"
    )
    parser.add_argument(
        "--gnuplot",
        dest="is_gnuplot",
        action="store_true",
        help="Output in gnuplot data style",
    )
    parser.add_argument(
        "-i",
        "--indices",
        dest="pdos_indices",
        help="Indices like 1 2, 3 4 5 6... (legacy plot only)",
    )
    parser.add_argument(
        "--legend", dest="show_legend", action="store_true", help="Show legend"
    )
    parser.add_argument(
        "--legacy",
        dest="is_legacy_plot",
        action="store_true",
        help="Plot in legacy style",
    )
    parser.add_argument(
        "--line",
        "-l",
        dest="is_vertical_line",
        action="store_true",
        help="Vertical line is drawn at between paths (legacy plot only)",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output_filename",
        action="store",
        help="Output filename of PDF plot",
    )
    parser.add_argument(
        "--xlabel", dest="xlabel", help="Specify x-label (legacy plot only)"
    )
    parser.add_argument(
        "--ylabel", dest="ylabel", help="Specify y-label (legacy plot only)"
    )
    parser.add_argument(
        "--points",
        dest="points",
        help="Draw points (o, '*', v, ^, x, p, d etc) (legacy plot only)",
    )
    parser.add_argument(
        "-t", "--title", dest="title", help="Title of plot (legacy plot only)"
    )
    parser.add_argument(
        "filenames",
        nargs="*",
        help=(
            "Filenames of phonon band structure result: band.yaml "
            "or band.hdf5(with --hdf5)"
        ),
    )
    args = parser.parse_args()
    return args


def _old_plot(args):
    import matplotlib.pyplot as plt

    if args.dos:
        import matplotlib.gridspec as gridspec

        plt.figure(figsize=(10, 6))
        gs = gridspec.GridSpec(1, 2, width_ratios=[3, 1])
        ax1 = plt.subplot(gs[0, 0])
        ax1.xaxis.set_ticks_position("both")
        ax1.yaxis.set_ticks_position("both")
        ax1.xaxis.set_tick_params(which="both", direction="in")
        ax1.yaxis.set_tick_params(which="both", direction="in")
    else:
        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")

    colors = [
        "r-",
        "b-",
        "g-",
        "c-",
        "m-",
        "y-",
        "k-",
        "r--",
        "b--",
        "g--",
        "c--",
        "m--",
        "y--",
        "k--",
    ]

    if args.points is None:
        marker = None
    else:
        marker = args.points

    if args.is_hdf5:
        if len(args.filenames) == 0:
            filenames = [
                "band.hdf5",
            ]
        else:
            filenames = args.filenames

    else:
        if len(args.filenames) == 0:
            filenames = ["band.yaml"]
        else:
            filenames = args.filenames

    if args.dos:
        dos_frequencies, dos = _read_dos_dat(
            args.dos, pdos_indices=args.pdos_indices, dos_factor=args.dos_factor
        )

    curves = []
    for i, filename in enumerate(filenames):
        if args.is_hdf5:
            (
                distances,
                frequencies,
                qpoints,
                segment_nqpoint,
                labels,
            ) = _read_band_hdf5(filename)
        else:
            (
                distances,
                frequencies,
                qpoints,
                segment_nqpoint,
                labels,
            ) = _read_band_yaml(filename)

        end_points = [
            0,
        ]
        for nq in segment_nqpoint:
            end_points.append(nq + end_points[-1])
        end_points[-1] -= 1
        segment_positions = distances[end_points]

        if not labels:
            labels_at_ends = None
        elif isinstance(labels[0], list):
            labels_at_ends = [
                labels[0][0],
            ]
            for j, pair in enumerate(labels[1:]):
                if labels[j][1] != pair[0]:
                    labels_at_ends.append("|".join([labels[j][1], pair[0]]))
                else:
                    labels_at_ends.append(pair[0])
            labels_at_ends.append(labels[-1][1])
        else:
            labels_at_ends = [labels[n] for n in end_points]

        if args.is_vertical_line and len(filenames) == 1:
            for v in segment_positions[1:-1]:
                plt.axvline(x=v, linewidth=0.5, color="b")

        q = 0
        for j, nq in enumerate(segment_nqpoint):
            if j == 0:
                curves.append(
                    plt.plot(
                        distances[q : (q + nq)],
                        frequencies[q : (q + nq)] * args.factor,
                        color=colors[i % len(colors)][0],
                        linestyle=colors[i % len(colors)][1:],
                        marker=marker,
                        label=filename,
                    )[0]
                )
            else:
                plt.plot(
                    distances[q : (q + nq)],
                    frequencies[q : (q + nq)] * args.factor,
                    color=colors[i % len(colors)][0],
                    marker=marker,
                    linestyle=colors[i % len(colors)][1:],
                )
            q += nq

    if args.xlabel is None:
        plt.xlabel("Wave vector")
    else:
        plt.xlabel(args.xlabel)
    if args.ylabel is None:
        plt.ylabel("Frequency")
    else:
        plt.ylabel(args.ylabel)

    plt.xlim(distances[0], distances[-1])
    if args.f_max is not None:
        plt.ylim(top=args.f_max)
    if args.f_min is not None:
        plt.ylim(bottom=args.f_min)
    plt.axhline(y=0, linestyle=":", linewidth=0.5, color="b")
    if len(filenames) == 1:
        xticks = segment_positions
        if args.band_labels:
            band_labels = [x for x in args.band_labels.split()]
            if len(band_labels) == len(xticks):
                plt.xticks(xticks, band_labels)
            else:
                print("Numbers of labels and band segments don't match.")
                sys.exit(1)
        elif labels_at_ends:
            plt.xticks(xticks, labels_at_ends)
        else:
            plt.xticks(xticks, [""] * len(xticks))
    else:
        plt.xticks([])

    if args.title is not None:
        plt.title(args.title)

    if args.show_legend:
        plt.legend(handles=curves)

    if args.dos:
        arg_fmax = len(dos_frequencies)
        if args.f_max is not None:
            for i, f in enumerate(dos_frequencies):
                if f > args.f_max:
                    arg_fmax = i
                    break
        arg_fmin = 0
        if args.f_min is not None:
            for i, f in enumerate(dos_frequencies):
                if f > args.f_min:
                    if i > 0:
                        arg_fmin = i - 1
                    break
        ax2 = plt.subplot(gs[0, 1], sharey=ax1)
        ax2.xaxis.set_ticks_position("both")
        ax2.yaxis.set_ticks_position("both")
        ax2.xaxis.set_tick_params(which="both", direction="in")
        ax2.yaxis.set_tick_params(which="both", direction="in")

        plt.subplots_adjust(wspace=0.03)
        plt.setp(ax2.get_yticklabels(), visible=False)

        for pdos in dos.T:
            if args.dos_max:
                _pdos = _get_dos(
                    pdos[arg_fmin:arg_fmax],
                    dos_frequencies[arg_fmin:arg_fmax],
                    args.dos_max,
                )
                plt.plot(_pdos[1], _pdos[0])
            else:
                plt.plot(pdos[arg_fmin:arg_fmax], dos_frequencies[arg_fmin:arg_fmax])
        plt.xlabel("DOS")

        ax2.set_xlim((0, None))

        if args.f_max is not None:
            plt.ylim(top=args.f_max)
        if args.f_min is not None:
            plt.ylim(bottom=args.f_min)
        if args.dos_max is not None:
            plt.xlim(right=args.dos_max)
        if args.dos_min is not None:
            plt.xlim(left=args.dos_min)

    if args.output_filename is not None:
        _savefig(plt, args.output_filename)
    else:
        plt.show()


def _plot(args):
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import ImageGrid

    if args.is_hdf5:
        if len(args.filenames) == 0:
            filenames = [
                "band.hdf5",
            ]
        else:
            filenames = args.filenames

        bands_data = [_read_band_hdf5(fname) for fname in filenames]
    else:
        if len(args.filenames) == 0:
            filenames = [
                "band.yaml",
            ]
        else:
            filenames = args.filenames
        bands_data = [_read_band_yaml(fname) for fname in filenames]

    if args.dos:
        dos_frequencies, dos = _read_dos_dat(
            args.dos, pdos_indices=args.pdos_indices, dos_factor=args.dos_factor
        )

    plots_data = [_arrange_band_data(*band_data) for band_data in bands_data]
    # Check consistency of input band structures
    all_path_connections = [data[1] for data in plots_data]
    wrong_file_i = _find_wrong_path_connections(all_path_connections)
    if wrong_file_i > 0:
        raise RuntimeError(
            "Band path of %s is inconsistent with %s."
            % (filenames[wrong_file_i], filenames[0])
        )

    # Decoration of figure
    max_frequencies = [_get_max_frequency(data[2]) for data in plots_data]
    plot_data = plots_data[np.argmax(max_frequencies)]
    _, path_connections, _, _ = plot_data
    n = len([x for x in path_connections if not x])

    if args.dos:
        n += 1

    fig = plt.figure()

    axs = ImageGrid(
        fig,
        111,  # similar to subplot(111)
        nrows_ncols=(1, n),
        axes_pad=0.11,
        label_mode="L",
    )
    for ax in axs[:-1]:
        if args.f_min:
            ax.set_ylim(ymin=args.f_min)
        if args.f_max:
            ax.set_ylim(ymax=args.f_max)

    band_plot = BandPlot(axs)
    band_plot.set_xscale_from_data(plot_data[2], plot_data[3])
    band_plot.xscale = band_plot.xscale * args.factor
    band_plot.decorate(*plot_data)

    # Plot band structures
    fmts = [
        "r-",
        "b-",
        "g-",
        "c-",
        "m-",
        "y-",
        "k-",
        "r--",
        "b--",
        "g--",
        "c--",
        "m--",
        "y--",
        "k--",
    ]
    for i, label in enumerate(filenames):
        _, p, f, d = plots_data[i]
        fmt = fmts[i % len(fmts)]
        _f = [f_seg * args.factor for f_seg in f]
        if args.show_legend:
            band_plot.plot(d, _f, p, fmt=fmt, label=_get_label_for_latex(label))
        else:
            band_plot.plot(d, _f, p, fmt=fmt)

    # dos
    if args.dos:
        arg_fmax = len(dos_frequencies)
        if args.f_max is not None:
            for i, f in enumerate(dos_frequencies):
                if f > args.f_max:
                    arg_fmax = i
                    break
        arg_fmin = 0
        if args.f_min is not None:
            for i, f in enumerate(dos_frequencies):
                if f > args.f_min:
                    if i > 0:
                        arg_fmin = i - 1
                    break

        axs[-1].xaxis.set_ticks_position("both")
        axs[-1].yaxis.set_ticks_position("both")
        axs[-1].xaxis.set_tick_params(which="both", direction="in")
        axs[-1].yaxis.set_tick_params(which="both", direction="in")

        for pdos in dos.T:
            if args.dos_max:
                _pdos = _get_dos(
                    pdos[arg_fmin:arg_fmax],
                    dos_frequencies[arg_fmin:arg_fmax],
                    args.dos_max,
                )
                axs[-1].plot(_pdos[1], _pdos[0])
            else:
                axs[-1].plot(
                    pdos[arg_fmin:arg_fmax], dos_frequencies[arg_fmin:arg_fmax]
                )
        axs[-1].set_xlabel("DOS")

        xlim = axs[-1].get_xlim()
        ylim = axs[-1].get_ylim()
        aspect = (xlim[1] - xlim[0]) / (ylim[1] - ylim[0]) * 3
        axs[-1].set_aspect(aspect)

    # Bring legend in front.
    if args.show_legend:
        axs[0].set_zorder(1)

    if args.title is not None:
        plt.suptitle(args.title)

    if args.output_filename is not None:
        _savefig(plt, args.output_filename)
    else:
        plt.show()


def _write_gnuplot_data(args):
    if args.is_hdf5:
        if len(args.filenames) == 0:
            filenames = [
                "band.hdf5",
            ]
        else:
            filenames = args.filenames

        bands_data = [_read_band_hdf5(fname) for fname in filenames]
    else:
        if len(args.filenames) == 0:
            filenames = [
                "band.yaml",
            ]
        else:
            filenames = args.filenames
        bands_data = [_read_band_yaml(fname) for fname in filenames]

    if args.is_gnuplot:
        distances = bands_data[0][0]
        frequencies = bands_data[0][1]
        segment_nqpoint = bands_data[0][3]

        end_points = [
            0,
        ]
        for nq in segment_nqpoint:
            end_points.append(nq + end_points[-1])
        end_points[-1] -= 1
        segment_positions = distances[end_points]

        print("# End points of segments: ")
        print("#   " + "%10.8f " * len(segment_positions) % tuple(segment_positions))
        for freqs in frequencies.T:
            q = 0
            for nq in segment_nqpoint:
                for d, f in zip(
                    distances[q : (q + nq)], freqs[q : (q + nq)] * args.factor
                ):
                    print("%f %f" % (d, f))
                q += nq
                print("")
            print("")


def run():
    """Run phonopy-bandplot."""
    args = get_options()
    if args.is_gnuplot:
        _write_gnuplot_data(args)
        sys.exit(1)

    if args.output_filename:
        import matplotlib

        matplotlib.use("Agg")

    if args.is_legacy_plot:
        _old_plot(args)
    else:
        _plot(args)



"""Tests for CP2K calculator interface."""

import os

import numpy as np
import pytest

from phonopy.interface.cp2k import read_cp2k
from phonopy.interface.phonopy_yaml import read_cell_yaml

data_dir = os.path.dirname(os.path.abspath(__file__))

CP2K_INPUT_TOOLS_AVAILABLE = True

try:
    import cp2k_input_tools  # noqa F401
except ImportError:
    CP2K_INPUT_TOOLS_AVAILABLE = False


@pytest.mark.skipif(
    not CP2K_INPUT_TOOLS_AVAILABLE, reason="not found cp2k-input-tools package"
)
def test_read_cp2k():
    """Test read_cp2k."""
    cell, _ = read_cp2k(os.path.join(data_dir, "Si-CP2K.inp"))
    cell_ref = read_cell_yaml(os.path.join(data_dir, "Si-CP2K.yaml"))
    assert (np.abs(cell.cell - cell_ref.cell) < 1e-5).all()
    diff_pos = cell.scaled_positions - cell_ref.scaled_positions
    diff_pos -= np.rint(diff_pos)
    assert (np.abs(diff_pos) < 1e-5).all()
    for s, s_r in zip(cell.symbols, cell_ref.symbols):
        assert s == s_r



"""Phonon calculation on sampling mesh."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np

from phonopy.harmonic.dynamical_matrix import (
    DynamicalMatrix,
    run_dynamical_matrix_solver_c,
)
from phonopy.physical_units import get_physical_units
from phonopy.structure.grid_points import GridPoints


class MeshBase:
    """Base class of Mesh and IterMesh classes.

    Attributes
    ----------
    mesh_numbers: ndarray
        Mesh numbers along a, b, c axes.
        dtype='intc'
        shape=(3,)
    qpoints: ndarray
        q-points in reduced coordinates of reciprocal lattice
        dtype='double'
        shape=(ir-grid points, 3)
    weights: ndarray
        Geometric q-point weights. Its sum is the number of grid points.
        dtype='intc'
        shape=(ir-grid points,)
    grid_address: ndarray
        Addresses of all grid points represented by integers.
        dtype='intc'
        shape=(prod(mesh_numbers), 3)
    ir_grid_points: ndarray
        Indices of irreducibple grid points in grid_address.
        dtype='intc'
        shape=(ir-grid points,)
    grid_mapping_table: ndarray
        Index mapping table from all grid points to ir-grid points.
        dtype='intc'
        shape=(prod(mesh_numbers),)
    dynamical_matrix: DynamicalMatrix
        Dynamical matrix instance to compute dynamical matrix at q-points.

    """

    def __init__(
        self,
        dynamical_matrix: DynamicalMatrix,
        mesh,
        shift=None,
        is_time_reversal=True,
        is_mesh_symmetry=True,
        with_eigenvectors=False,
        is_gamma_center=False,
        rotations=None,  # Point group operations in real space
        factor=None,
    ):
        """Init method."""
        self._mesh = np.array(mesh, dtype="intc")
        self._with_eigenvectors = with_eigenvectors
        if factor is None:
            self._factor = get_physical_units().DefaultToTHz
        else:
            self._factor = factor
        self._cell = dynamical_matrix.primitive
        self._dynamical_matrix = dynamical_matrix

        self._gp = GridPoints(
            self._mesh,
            np.linalg.inv(self._cell.cell),
            q_mesh_shift=shift,
            is_gamma_center=is_gamma_center,
            is_time_reversal=(is_time_reversal and is_mesh_symmetry),
            rotations=rotations,
            is_mesh_symmetry=is_mesh_symmetry,
        )

        self._qpoints = self._gp.qpoints
        self._weights = self._gp.weights

        self._frequencies = None
        self._eigenvectors = None

        self._q_count = 0

    @property
    def mesh_numbers(self):
        """Return mesh numbers."""
        return self._mesh

    def get_mesh_numbers(self):
        """Return mesh numbers."""
        warnings.warn(
            "MeshBase.get_mesh_numbers() is deprecated. Use mesh_numbers attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.mesh_numbers

    @property
    def qpoints(self):
        """Return (irreducible) q-points."""
        return self._qpoints

    def get_qpoints(self):
        """Return (irreducible) q-points."""
        warnings.warn(
            "MeshBase.get_qpoints() is deprecated. Use qpoints attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.qpoints

    @property
    def weights(self):
        """Return (irreducible) weights of q-points."""
        return self._weights

    def get_weights(self):
        """Return (irreducible) weights of q-points."""
        warnings.warn(
            "MeshBase.get_weights()) is deprecated. Use weights attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.weights

    @property
    def grid_address(self):
        """Return mesh grid addresses."""
        return self._gp.grid_address

    def get_grid_address(self):
        """Return mesh grid addresses."""
        warnings.warn(
            "MeshBase.get_grid_address()) is deprecated. Use grid_address attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.grid_address

    @property
    def ir_grid_points(self):
        """Return irreducible grid indices."""
        return self._gp.ir_grid_points

    def get_ir_grid_points(self):
        """Return irreducible grid indices."""
        warnings.warn(
            "MeshBase.get_ir_grid_points() is deprecated. "
            "Use ir_grid_points attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.ir_grid_points

    @property
    def grid_mapping_table(self):
        """Return grid index mapping table."""
        return self._gp.grid_mapping_table

    def get_grid_mapping_table(self):
        """Return grid index mapping table."""
        warnings.warn(
            "MeshBase.get_grid_mapping_table() is deprecated. "
            "Use grid_mapping_table attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.grid_mapping_table

    @property
    def dynamical_matrix(self):
        """Return dynamical matrix class instance."""
        return self._dynamical_matrix

    def get_dynamical_matrix(self):
        """Return dynamical matrix class instance."""
        warnings.warn(
            "MeshBase.get_dynamical_matrix() is deprecated. "
            "Use dynamical_matrix attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.dynamical_matrix

    @property
    def with_eigenvectors(self):
        """Whether eigenvectors are calculated or not."""
        return self._with_eigenvectors


class Mesh(MeshBase):
    """Class for phonons on mesh grid.

    Frequencies and eigenvectors can be also accessible by iterator
    representation to be compatible with IterMesh.

    Attributes
    ----------
    frequencies: ndarray
        Phonon frequencies at ir-grid points. Imaginary frequenies are
        represented by negative real numbers.
        dtype='double'
        shape=(ir-grid points, bands)
    eigenvectors: ndarray
        Phonon eigenvectors at ir-grid points. See the data structure at
        np.linalg.eigh.
        shape=(ir-grid points, bands, bands)
        dtype=complex of "c%d" % (np.dtype('double').itemsize * 2)
        order='C'
    group_velocities: ndarray
        Phonon group velocities at ir-grid points.
        shape=(ir-grid points, bands, 3)
        dtype='double'
    More attributes from MeshBase should be watched.

    """

    def __init__(
        self,
        dynamical_matrix,
        mesh,
        shift=None,
        is_time_reversal=True,
        is_mesh_symmetry=True,
        with_eigenvectors=False,
        is_gamma_center=False,
        group_velocity=None,
        rotations=None,  # Point group operations in real space
        factor=None,
    ):
        """Init method."""
        super().__init__(
            dynamical_matrix,
            mesh,
            shift=shift,
            is_time_reversal=is_time_reversal,
            is_mesh_symmetry=is_mesh_symmetry,
            with_eigenvectors=with_eigenvectors,
            is_gamma_center=is_gamma_center,
            rotations=rotations,
            factor=factor,
        )

        self._group_velocity = group_velocity
        self._group_velocities = None

    def __iter__(self):
        """Define iterator over q-points.

        Initially, all phonons are computed and stored in arrays.
        Then this is just used as an iterator to return exisiting results.
        The purpose of this iterator is compatible use of IterMesh.

        """
        if self._frequencies is None:
            self.run()
        return self

    def __next__(self):
        """Return phonon frequencies and eigenvectors at each q-point."""
        if self._q_count == len(self._qpoints):
            self._q_count = 0
            raise StopIteration
        else:
            i = self._q_count
            self._q_count += 1
            if self._eigenvectors is None:
                return self._frequencies[i], None
            else:
                return self._frequencies[i], self._eigenvectors[i]

    def run(self):
        """Calculate phonons at all required q-points."""
        self._set_phonon()
        if self._group_velocity is not None:
            self._set_group_velocities(self._group_velocity)

    @property
    def frequencies(self):
        """Return phonon frequencies."""
        if self._frequencies is None:
            self.run()
        return self._frequencies

    def get_frequencies(self):
        """Return phonon frequencies."""
        warnings.warn(
            "Mesh.get_frequencies() is deprecated. Use frequencies attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.frequencies

    @property
    def eigenvectors(self):
        """Return eigenvectors.

        Eigenvectors is a numpy array of three dimension.
        The first index runs through q-points.
        In the second and third indices, eigenvectors obtained
        using numpy.linalg.eigh are stored.

        The third index corresponds to the eigenvalue's index.
        The second index is for atoms [x1, y1, z1, x2, y2, z2, ...].

        """
        if self._frequencies is None:
            self.run()
        return self._eigenvectors

    def get_eigenvectors(self):
        """Return eigenvectors."""
        warnings.warn(
            "Mesh.get_eigenvectors() is deprecated. Use eigenvectors attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.eigenvectors

    @property
    def group_velocities(self):
        """Return group velocities."""
        if self._frequencies is None:
            self.run()
        return self._group_velocities

    def get_group_velocities(self):
        """Return group velocities."""
        warnings.warn(
            "Mesh.get_group_velocities() is deprecated. "
            "Use group_velocities attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.group_velocities

    def write_hdf5(self, filename="mesh.hdf5"):
        """Write results to hdf5 file."""
        import h5py

        with h5py.File(filename, "w") as w:
            w.create_dataset("mesh", data=self._mesh)
            w.create_dataset("qpoint", data=self._qpoints)
            w.create_dataset("weight", data=self._weights)
            w.create_dataset("frequency", data=self._frequencies)
            if self._eigenvectors is not None:
                w.create_dataset("eigenvector", data=self._eigenvectors)
            if self._group_velocities is not None:
                w.create_dataset("group_velocity", data=self._group_velocities)

    def write_yaml(self, filename="mesh.yaml"):
        """Write results to yaml file."""
        natom = len(self._cell)
        rec_lattice = np.linalg.inv(self._cell.cell)  # column vectors
        distances = np.sqrt(np.sum(np.dot(self._qpoints, rec_lattice.T) ** 2, axis=1))

        lines = []

        lines.append("mesh: [ %5d, %5d, %5d ]" % tuple(self._mesh))
        lines.append("nqpoint: %-7d" % self._qpoints.shape[0])
        lines.append("reciprocal_lattice:")
        for vec, axis in zip(rec_lattice.T, ("a*", "b*", "c*")):
            lines.append("- [ %12.8f, %12.8f, %12.8f ] # %2s" % (tuple(vec) + (axis,)))
        lines.append("natom:   %-7d" % natom)
        lines.append(str(self._cell))
        lines.append("")
        lines.append("phonon:")

        for i, (q, d) in enumerate(zip(self._qpoints, distances)):
            lines.append("- q-position: [ %12.7f, %12.7f, %12.7f ]" % tuple(q))
            lines.append("  distance_from_gamma: %12.9f" % d)
            lines.append("  weight: %-5d" % self._weights[i])
            lines.append("  band:")

            for j, freq in enumerate(self._frequencies[i]):
                lines.append("  - # %d" % (j + 1))
                lines.append("    frequency:  %15.10f" % freq)

                if self._group_velocities is not None:
                    lines.append(
                        "    group_velocity: "
                        "[ %13.7f, %13.7f, %13.7f ]"
                        % tuple(self._group_velocities[i, j])
                    )

                if self._with_eigenvectors:
                    lines.append("    eigenvector:")
                    for k in range(natom):
                        lines.append("    - # atom %d" % (k + 1))
                        for ll in (0, 1, 2):
                            lines.append(
                                "      - [ %17.14f, %17.14f ]"
                                % (
                                    self._eigenvectors[i, k * 3 + ll, j].real,
                                    self._eigenvectors[i, k * 3 + ll, j].imag,
                                )
                            )
            lines.append("")

        with open(filename, "w") as w:
            w.write("\n".join(lines))

    def _set_phonon(self):
        import phonopy._phonopy as phonoc

        num_band = len(self._cell) * 3
        num_qpoints = len(self._qpoints)

        self._frequencies = np.zeros((num_qpoints, num_band), dtype="double")
        if phonoc.use_openmp():
            dynmat = run_dynamical_matrix_solver_c(
                self._dynamical_matrix, self._qpoints
            )
            eigenvectors = dynmat
        elif self._with_eigenvectors:
            dtype = "c%d" % (np.dtype("double").itemsize * 2)
            eigenvectors = np.zeros(
                (
                    num_qpoints,
                    num_band,
                    num_band,
                ),
                dtype=dtype,
                order="C",
            )

        for i, q in enumerate(self._qpoints):
            if phonoc.use_openmp():
                dm = dynmat[i]
            else:
                self._dynamical_matrix.run(q)
                dm = self._dynamical_matrix.dynamical_matrix
            if self._with_eigenvectors:
                eigvals, eigenvectors[i] = np.linalg.eigh(dm)
                eigenvalues = eigvals.real
            else:
                eigenvalues = np.linalg.eigvalsh(dm).real
            self._frequencies[i] = (
                np.array(
                    np.sqrt(abs(eigenvalues)) * np.sign(eigenvalues),
                    dtype="double",
                    order="C",
                )
                * self._factor
            )

        if self._with_eigenvectors:
            self._eigenvectors = eigenvectors

    def _set_group_velocities(self, group_velocity):
        group_velocity.run(self._qpoints)
        self._group_velocities = group_velocity.group_velocities


class IterMesh(MeshBase):
    """Generator class for phonons on mesh grid.

    Not like as Mesh class, frequencies and eigenvectors are not
    stored, instead generated by iterator. This may be used for
    saving memory space even with very dense samplig mesh.

    Attributes
    ----------
    Attributes from MeshBase should be watched.

    """

    def __init__(
        self,
        dynamical_matrix,
        mesh,
        shift=None,
        is_time_reversal=True,
        is_mesh_symmetry=True,
        with_eigenvectors=False,
        is_gamma_center=False,
        rotations=None,  # Point group operations in real space
        factor=None,
    ):
        """Init method."""
        super().__init__(
            dynamical_matrix,
            mesh,
            shift=shift,
            is_time_reversal=is_time_reversal,
            is_mesh_symmetry=is_mesh_symmetry,
            with_eigenvectors=with_eigenvectors,
            is_gamma_center=is_gamma_center,
            rotations=rotations,
            factor=factor,
        )

    def __iter__(self):
        """Define iterator over q-points."""
        return self

    def __next__(self):
        """Calculate phonons at a q-point."""
        if self._q_count == len(self._qpoints):
            self._q_count = 0
            raise StopIteration
        else:
            q = self._qpoints[self._q_count]
            self._dynamical_matrix.run(q)
            dm = self._dynamical_matrix.dynamical_matrix
            if self._with_eigenvectors:
                eigvals, eigenvectors = np.linalg.eigh(dm)
                eigenvalues = eigvals.real
            else:
                eigenvalues = np.linalg.eigvalsh(dm).real
            frequencies = (
                np.array(
                    np.sqrt(abs(eigenvalues)) * np.sign(eigenvalues),
                    dtype="double",
                    order="C",
                )
                * self._factor
            )
            self._q_count += 1
            return frequencies, eigenvectors



"""Tests of Phonopy --symmetry."""

from __future__ import annotations

import os
import pathlib
from collections.abc import Sequence
from dataclasses import dataclass, fields
from typing import Optional, Union

import numpy as np
import pytest

from phonopy.cui.phonopy_script import main

cwd = pathlib.Path(__file__).parent
cwd_called = pathlib.Path.cwd()


@dataclass
class MockArgs:
    """Mock args of ArgumentParser."""

    filename: Optional[Sequence[Union[os.PathLike, str]]] = None
    conf_filename: Optional[os.PathLike] = None
    log_level: Optional[int] = None
    fc_symmetry: bool = True
    cell_filename: Optional[str] = None
    conf_filename: Optional[str] = None
    create_force_sets: Optional[list[str]] = None
    is_check_symmetry: Optional[bool] = None
    is_graph_plot: Optional[bool] = None
    is_graph_save: Optional[bool] = None
    is_legend: Optional[bool] = None
    is_displacement: Optional[bool] = None
    supercell_dimension: Optional[str] = None
    magmoms: Optional[str] = None
    use_pypolymlp: bool = False

    def __iter__(self):
        """Make self iterable to support in."""
        return (getattr(self, field.name) for field in fields(self))

    def __contains__(self, item):
        """Implement in operator."""
        return item in (field.name for field in fields(self))


@pytest.mark.parametrize("is_ncl", [False, True])
def test_phonopy_disp_Cr(is_ncl: bool):
    """Test phonopy -d option."""
    cell_filename = cwd / "POSCAR-unitcell_Cr"
    if is_ncl:
        magmoms = "0 0 1 0 0 -1"
    else:
        magmoms = "1 -1"

    argparse_control = _get_phonopy_args(
        cell_filename=cell_filename,
        supercell_dimension="2 2 2",
        is_displacement=True,
        magmoms=magmoms,
    )
    with pytest.raises(SystemExit) as excinfo:
        main(**argparse_control)
    assert excinfo.value.code == 0

    magmom_file_path = pathlib.Path(cwd_called / "MAGMOM")
    assert magmom_file_path.exists()
    with open(magmom_file_path) as f:
        vals = [float(v) for v in f.readline().split()[2:]]

    if is_ncl:
        np.testing.assert_allclose(vals, np.ravel([[0, 0, 1]] * 8 + [[0, 0, -1]] * 8))
    else:
        np.testing.assert_allclose(vals, [1.0] * 8 + [-1.0] * 8)

    for created_filename in ["MAGMOM", "POSCAR-001", "SPOSCAR", "phonopy_disp.yaml"]:
        file_path = pathlib.Path(cwd_called / created_filename)
        assert file_path.exists()
        file_path.unlink()


def test_create_force_sets():
    """Test phonopy --create-force-sets command."""
    argparse_control = _get_phonopy_args(
        cell_filename=cwd / "NaCl" / "phonopy_disp.yaml.xz",
        create_force_sets=[
            cwd / "NaCl" / "vasprun.xml-001.xz",
            cwd / "NaCl" / "vasprun.xml-002.xz",
        ],
        load_phonopy_yaml=False,
    )
    with pytest.raises(SystemExit) as excinfo:
        main(**argparse_control)
    assert excinfo.value.code == 0

    for created_filename in ("FORCE_SETS",):
        file_path = pathlib.Path(cwd_called / created_filename)
        assert file_path.exists()
        file_path.unlink()

    argparse_control = _get_phonopy_args(
        cell_filename=cwd / "NaCl" / "phonopy_disp.yaml.xz",
        create_force_sets=[
            cwd / "NaCl" / "vasprun.xml-002.xz",
            cwd / "NaCl" / "vasprun.xml-001.xz",
        ],
    )
    with pytest.raises(RuntimeError) as excinfo:
        main(**argparse_control)


@pytest.mark.parametrize("load_phonopy_yaml", [False, True])
def test_phonopy_load(load_phonopy_yaml: bool):
    """Test phonopy-load command."""
    pytest.importorskip("symfc")
    # Check sys.exit(0)
    argparse_control = _get_phonopy_args(
        filename=cwd / ".." / "phonopy_params_NaCl-1.00.yaml.xz",
        load_phonopy_yaml=load_phonopy_yaml,
    )
    with pytest.raises(SystemExit) as excinfo:
        main(**argparse_control)
    assert excinfo.value.code == 0

    # Clean files created by phonopy-load script.
    for created_filename in ("phonopy.yaml",):
        file_path = pathlib.Path(cwd_called / created_filename)
        assert file_path.exists()
        file_path.unlink()


def test_phonopy_is_check_symmetry():
    """Test phonopy --symmetry command with phonopy.yaml input structure."""
    # Check sys.exit(0)
    argparse_control = _get_phonopy_args(
        filename=cwd / ".." / "phonopy_params_NaCl-1.00.yaml.xz",
        load_phonopy_yaml=False,
        is_check_symmetry=True,
    )
    with pytest.raises(SystemExit) as excinfo:
        main(**argparse_control)
    assert excinfo.value.code == 0

    # Clean files created by phonopy --symmetry command.
    file_path = pathlib.Path(cwd_called / "phonopy_symcells.yaml")
    assert file_path.exists()
    file_path.unlink()


def test_conf_file():
    """Test phonopy CONFILE."""
    argparse_control = _get_phonopy_args(
        filename=cwd / "dim.conf",
        cell_filename=cwd / "POSCAR-unitcell_Cr",
        load_phonopy_yaml=False,
    )
    with pytest.raises(SystemExit) as excinfo:
        main(**argparse_control)
    assert excinfo.value.code == 0

    for created_filename in ("POSCAR-001", "SPOSCAR", "phonopy_disp.yaml"):
        file_path = pathlib.Path(cwd_called / created_filename)
        assert file_path.exists()
        file_path.unlink()


def test_config_option():
    """Test phonopy-yaml --config."""
    pytest.importorskip("symfc")
    argparse_control = _get_phonopy_args(
        filename=cwd / ".." / "phonopy_params_NaCl-1.00.yaml.xz",
        conf_filename=cwd / "mesh.conf",
        load_phonopy_yaml=True,
    )
    with pytest.raises(SystemExit) as excinfo:
        main(**argparse_control)
    assert excinfo.value.code == 0

    # Clean files created by phonopy-load script.
    for created_filename in ("phonopy.yaml", "mesh.yaml"):
        file_path = pathlib.Path(cwd_called / created_filename)
        assert file_path.exists()
        file_path.unlink()


def _get_phonopy_args(
    cell_filename: Optional[Union[str, pathlib.Path]] = None,
    create_force_sets: Optional[list[str]] = None,
    supercell_dimension: Optional[str] = None,
    is_displacement: Optional[bool] = None,
    magmoms: Optional[str] = None,
    load_phonopy_yaml: bool = False,
    is_check_symmetry: bool = False,
    filename: Optional[str] = None,
    conf_filename: Optional[str] = None,
    use_pypolymlp: bool = False,
):
    if filename is None:
        _filename = []
    else:
        _filename = [filename]
    mockargs = MockArgs(
        filename=_filename,
        log_level=1,
        cell_filename=cell_filename,
        create_force_sets=create_force_sets,
        supercell_dimension=supercell_dimension,
        is_displacement=is_displacement,
        magmoms=magmoms,
        is_check_symmetry=is_check_symmetry,
        conf_filename=conf_filename,
        use_pypolymlp=use_pypolymlp,
    )

    if load_phonopy_yaml:
        argparse_control = {
            "fc_symmetry": True,
            "is_nac": True,
            "load_phonopy_yaml": True,
            "args": mockargs,
        }
    else:
        argparse_control = {
            "fc_symmetry": False,
            "is_nac": False,
            "load_phonopy_yaml": False,
            "args": mockargs,
        }
    return argparse_control
